/*----------------------------------------------------------------------------*/
/* app2.c - application dependent routines.  Example from Goldberg's book.    */
/*          Version of appxxxxx.c changed to set up solution of this problem. */
/* This example application, app2.c, interprets chromosomes as concatenated   */
/* strings of binary integers of user-specified length (field_size).  Fitness */
/* is simply the length of this integer vector squared.                       */
/* This example is particularly interesting because it illustrates the use    */
/* of the "utility" fields pointer on the chromosome.                         */
/* Note:  The way this example is encoded is taken from SGA-C, and does       */
/* NOT reflect the best way to encode it using the additional tools           */
/* provided by GALOPPS (which could use chromtointarray(), and alpha_size >2, */
/* to make the problem easier (but that would not allow for a "short" final   */
/* field, which this does.                                                    */
/*----------------------------------------------------------------------------*/

#include "external.h"

static int field_size, vec_size;



void
objfunc(critter)
/* Application dependent objective function */
struct individual *critter;
{
    int     i, start, stop;

    neval++;
    local_cycle_neval++;

    /* Interpret each chromosome as a vector of concatenated integers */

    critter->neval = neval;
    critter->init_fitness = 0.0;
    critter->fitness = 0.;

    for (i = 0; i < vec_size; i++) {
	/* section of chromosome containing current integer field */
	start = (i * field_size) + 1;
	stop = ((i + 1) * field_size);

	/* check if enough bits remain, if not, interpret as a short field */
	if (stop > lchrom)
	    stop = lchrom;
	/* NOTE:  In GALOPPS (as opposed to SGA-C), it would be easier to */
	/* represent this problem NOT with alpha_size = 2 (binary), but with */
	/* alpha_size at whatever power of 2 is desired, and then use */
	/* chromtointarray() to decode the chromosome to a local int array. */
	/* But the scheme above and below is from SGA-C. */
     
	/* convert bit field in chromosome to an integer    */
	/* and store it in utility array.  Then compute     */
	/* chromosome fitness as sum of squares of integers */
	critter->utility[i] = ithruj2int(start, stop, critter->chrom);
	critter->init_fitness += (float) critter->utility[i] *
	 critter->utility[i];
    }
}


void
app_set_options()
/* This routine is called once when a new run or extended run is started. */
/* It must set a value for all of its global variables, although they may */
/* be modified later (in other app_ functions) during the run. */
{

  avg_mut_width = 3;    /* Must set if using multimut.c.  Value is expected  */
			/* number of adjacent fields (bits, if binary) to    */
			/* change at once.  */

  stochastic = 0;       /* Must set to 0 or 1 to specify whether or not */
			/* fitness function always returns same value for a  */
			/* given chrom.-- If always same, set stochastic = 0 */
			/* to avoid extra evaluations; if changes with env't */
			/* or randomly, set = 1 */

  elitism = 1;          /* Must set to 0 or 1 to determine whether or not     */
			/* best indiv. is always preserved in next generation */
			/* (1 = yes).  */
  elitism &= !stochastic;   /* Note that elitism is reset to 0 automatically  */
			/* if user specifies stochastic flag = 1, since the   */
			/* changing fitness makes it impossible to insure that*/
			/* what was once best is still best. */

  user_supplied_initialization = 0;       /* Make it TRUE (1) if you will    */
					  /* supply code to initialize the   */
					  /* individuals in oldpop below, in */
					  /* routine app_user_init_pop(),    */
					  /* instead of using one of the     */
					  /* "standard" methods supplied with*/
					  /* GALOPPS.  */

  call_app_user_does_migration = 0;   /* Flag, if set, allows user full   */
					/* control of migration, from */
					/* app_user_does_migration(), below */

  using_inversion = 0;
			 /* Set flag to 1 (TRUE) if will use inversion to  */
			 /* alter placement of fields on chromosome, with  */
			 /* remapping of fields to "standard" positions at */
			 /* each evaluation, before objective function is  */
			 /* called.  It is set automatically if pinversion>0, */
			 /* but user may set manually here if has ever used */
			 /* inversion, so unmapping, translation, etc., will*/
			 /* continue to occur, even if pinversion now set  */
			 /* to 0.  */

  different_reps = 0 || using_inversion;
			 /* You set to 1 (TRUE) if will use different reps */
			 /* in different subpops.  If TRUE, and if the */
			 /* difference is NOT ONLY INVERSION, then YOU MUST*/
			 /* also supply code to transform ALL of the */
			 /* migrating individuals, in app_transform_migrant*/
			 /* in the skeleton below, and code below in */
			 /* app_transform_chrom_to_std to transform a */
			 /* chromosome into its STANDARD representation */
			 /* (used in the objective function objfunc().  */
			 /* Always 1 if inversion is being used.  */

  if (different_reps && !using_inversion) {
      /* User is using different representations in various subpopulations, */
      /* so must be supplying app_transform_chrom_to_std() and */
      /* app_transform_migrant().  User must also set the variables */
      /* max_chromsize_among_all_subpops and max_numfields_among_all_subpops */
      /* in the lines directly below, and delete the print and the call to */
      /* exit() which are inserted as a reminder to the user.  */ 

      fprintf(stderr, "\n*** You have selected different_reps and are not ");
      fprintf(stderr,"using inversion, \nso you MUST specify here in app_init");
      fprintf(stderr, " the maximum sizes of chromsize and numfields.***\n");
      exit(-9);
  }
  else {
      max_chromsize_among_all_subpops = -1; /* set to illegal value, as flag */
      max_numfields_among_all_subpops = -1; /* set to illegal value, as flag */
  }
}


void 
app_set_field_sizes()
/* Routine called ONLY IF user inputs alpha_size as < 2, meaning fields have */
/* different lengths.  In that case, user MUST provide code here to set the  */
/* array field_sizes[i], 0<=i<numfields, to the number of different codes    */
/* possible in the ith field (codes will vary from 0 to field_sizes[i] - 1). */
/* Data may be set in executable statements, or via initializers.   If the   */
/* problem uses different representations in different subpopulations, and   */
/* if that affects the field sizes, then this code must set the proper sizes */
/* for each subpopulation, using popno to tell which subpop is being         */
/* initialized.  Elsewhere, the program will also automatically use the      */
/* field_sizes to set array  field_ends, to the final bit number (in         */
/* [1, lchrom]) of each field.   */
{
  int i;

      field_sizes[0] = 24;  /* Replace with code to set them to the sizes  */
			    /* you actually want.  */
      field_sizes[9] = 48;  /* Replace with code to set them to the sizes  */
}


void
app_read_prob_params()
/* Application dependent data input, called ONCE at start of run (whether */
/* started from input parameters or EXTENDED from restart files) by main */
/* program BEFORE it reads the "standard" fields like numfields, */
/* numextrafields, lchrom, etc.  */
/* (Note: Onepop may do many RUNS, and this is called at start of each). */ 
/* Values read here are usually stored in static variables */
/* declared at the top of the app file, so they may be used in any of the */
/* user callback routines, but do NOT have to be read/written to the */
/* checkpoint files.  Thus, user should read these values ALSO when a run is */
/* restarted from saved checkpoint files.  If values MUST be preserved ACROSS */
/* runs (i.e., saved when checkpointing and reread when restarting) then */
/* that must be done in app_write_ckp_hdr() and app_read_ckp_hdr(), below, */
/* and the variables will be saved with EACH subpopulation.  Note: values */
/* like numfields, alpha_size, etc., are NOT yet known when this is called. */
{
    int     size;

    if (extending) {
       /* Strange machinations are needed in THIS application to allow
runs to be EXTENDED from saved checkpoint files, because when the program 
is first started (from scratch), it determines the number of utility 
fields to make (for recording bestever utility, and utility fields of
each individual) in app_data, below, AFTER it already knows numfields,
fieldlength, lchrom, etc.  However, WHEN A RUN IS BEING EXTENDED from
saved checkpoint files, there is a problem:  app_data is NOT called (if
at all) until AFTER the malloc's of the utility fields to allow reading in 
of the checkpoint files is done, so this information must be supplied
somewhere BEFORE the checkpoint file is read (in THIS routine, for
example).  That's fine, but THIS routine is called before numfields, 
fieldlength, lchrom, etc., are restored, so we can't use the SAME calculation 
here as we use on a start-from-scratch run in app_data.  We must ask the
user to provide BOTH vec_size and field_size, since we can't use lchrom
to calculate one from the other.  So here it is:                          */


       /* user must specify LENGTH (bits) of each integer in chromosome. */
       if (interactive)
	  printf(" Enter field size used in this run -------->" 
		     , field_size);
       ffscanf("field_size", infp, "%d", &field_size);

       /* user must specify NUMBER of concatenated integers in chromosome. */
       if (interactive)
	  printf(" Enter number of those integer fields used in this run -->");
       ffscanf("vec_size", infp, "%d", &vec_size);

    }
}


void
app_data(REVISING)
BOOL    REVISING;
/* Application dependent data input, called by initdata() and revisedata(), */
/* just after user answers questions about crossover rate, scaling, */
/* convinterval, etc.  Ask input questions here which might be different for */
/* each subpopulation, and write (and read) the values to the checkpoint */
/* header file by putting code into app_write_ckp_hdr() and app_read_ckp_hdr()*/
/* below.  Values like numfields, alpha_size, etc., are already known when */
/* this routine is called. */
/* NOTE:  app_data is called whenever the user is initializing (from user */
/* input, flag REVISING == FALSE) or revising (flag REVISING == TRUE) input */
/* data (but only once if flag all_subpops_use_same_parameters is TRUE). */
/* HOWEVER, IF user is RESTARTING from saved checkpoint files, and does NOT */
/* ask to revise any parameters, this routing IS NOT CALLED.  Therefore, for */
/* user data (tables, etc.) which you want to read in on a RESTART, also, */
/* use app_init to read those data, as it is called whether restarting or */
/* starting from the beginning.  (Put it inside the "firstcall" logic.) */
{
    int     size = UINTSIZE;

    if (lchrom < UINTSIZE)
	size = lchrom;
    if (REVISING)
	return;
    /* user must specify length of concatenated integers in the chromosome. */
    if (interactive)
	printf(" Enter field size (must be less than %d) (%d)->", size
		     , field_size);
    ffscanf("field_size", infp, "%d", &field_size);
    vec_size = lchrom / field_size;
    if ((((float) lchrom / (float) field_size) - vec_size) > 0.0)
	vec_size++;

}


void
application()
/* this routine should contain any application-dependent computations */
/* that should be performed before each generation. called by main()    */
{
}


void
app_init()
/* Application dependent initialization routine called by initialize() for */
/* EACH subpopulation each cycle, whether started from input or restarted  */
/* from a restartfile, and whether all_subpops_use_same_parameters is TRUE */
/* or FALSE.  Called after individuals are malloc'd and after individuals  */
/* are read from .ind file (on a restart), but before random initialization*/
/* of all individuals (if initial start) or of individuals NOT read (if    */
/* restarting).                                                            */
{
  static int    firstcall = 1;

  /* Any code you want to add to app_init goes below. */

  if (firstcall) {
    firstcall = 0;
    /* Put here any code you want to do just once in a run, regardless of */
    /* how run was started. */

    return;
  }

  /* Put here any code you want run for each subpopulation at the beginning */
  /* of its execution in each cycle. */

  return;
}


void
app_user_init_pop(starting_guy_index)
/* If none of the standard methods supplied with GALOPPS will work for */
/* initializing the individuals in oldpop at start of run, or for creating */
/* new "random" guys when popsize is expanded, or when convergence triggers*/
/* partial re-initialization, etc., then you can create the individuals */
/* here, filling oldpop[starting_guy_index] through oldpop[popsize-1].  */
int  starting_guy_index;
{
}


void
app_after_random_init()
/* Application-dependent changes or redoing of initialization, called after */
/* the random initialization is performed and (on restart) individuals have */
/* been read in from restart file. */
{
}


void
app_initreport()
/* Application-dependent initial report called by initialize() */
{
    if (vec_size > lchrom / field_size) {
	fprintf(outfp, " Each chromosome interpreted as %d %d-bit integers",
		vec_size - 1, field_size);
	fprintf(outfp, " and one %d-bit integer.\n", lchrom - ((vec_size - 1)
							     * field_size));
    }
    else
	fprintf(outfp, "Each chromosome interpreted as %d %d-bit integers.\n",
		vec_size, field_size);
}



void
app_report()
/* Application-dependent report, called by report() */
{
    int     i, j;


    /* Print vector interpretation of each chromosome. */
    if (printstrings == 1) {
	for (i = 0; i < popsize; i++) {
	    fprintf(outfp, "individual %d = ", i+1); 
	    /* external numbers run 1,2,3,..., NOT 0,1,2,...  */
	    for (j = 0; j < vec_size; j++)           
		fprintf(outfp, "%d ", oldpop[i].utility[j]);
	    fprintf(outfp, "\n");
	}
    }
}


void
app_decide_if_converged(flag, num_to_replace_if_converged)
/* This routine is called at the end of each generation of each subpopulation */
/* to give the user the option of checking for convergence of the */
/* subpopulation, or other condition.  When (if) that happens, if user wants  */
/* simply to reinitialize all or part of the subpopulation to random values   */
/* and keep running, user can just set *flag to 1 and set */
/* num_to_replace_if_converged to the desired number to reinitialize.  If     */
/* user wants to take some other action, should set *flag = 2 and code the    */
/* desired actions in routine app_when_converged() below.  */
int * flag;
int * num_to_replace_if_converged;
{
}


void
app_when_converged()
/* This routine is called if user's routine (above) app_decide_if_converged  */
/* ever sets its flag to 2.  The intent is that user can here stop a run, or */
/* change any GA parameters as desired, for the current subpopulation, etc.  */
{
}


void
app_quiet_report()
 /* Routine to be called when other output is suppressed, to check for
  * trigger for special user-defined output, or to print desired output even
  * when running in quiet mode.  Helpful when doing many long runs.
  * User can do checking for last generation of a Manypops run, for example,
  * in case want to reset quiet to 0 for a final summary printout, etc.   */

{
    /* Below, supply any logic and printing to be used in quiet mode
     * operation, when most other normal output is suppressed. */

    /* for example, print warning ONE TIME if quiet flag is on, so user knows
     * why doesn't get any output.                                         */
    if(quiet < 3 && popno == 0 && cycle == 0 && gen == startgen)
	 fprintf(outfp, "\n Flag quiet > 0, normal output suppressed.\n");

    /* Logic below starts printing end of next-to-last generation of last
     * cycle, for final summary.  (Works for GALUPPS/Onepop (ncycles == 0) */
     /* and GALUPPS/Manypops) */
    if((cycle == ncycles - 1 || ncycles == 0) && gen == maxgen - 2) quiet = 0;

}

void
app_new_global_best_report()
/* Application-dependent report, called by globalstats() in statisti.c when  */
/* GALOPPS/Manypops is run (not used by GALOPPS/Onepop runs).  It is called  */
/* after the LAST GENERATION of the cycle of EACH subpopulation.  That means */
/* that some things MAY have changed (inversion pattern, etc.) in some       */
/* generations BETWEEN when this best individual was found and now, when it  */
/* is to be printed, so user must be aware of that in writing outputs.  Also */
/* called in checkhdr.c to print the same values (local_bestfit.xxx) when   */
/* that local best turns out also to be the best among all pops, all        */
/* processes, so is EQUIVALENT at that time to printing all_pops_bestfit.xxx).*/
/* It should print out any app-dependent fields user will want to see.      */
/* (This call is not made at all from Onepop -- single-population runs).    */
{
    /* User can print from utility fields, call output routines, etc.        */
    /* to print whatever information is wanted when a new best individual of */
    /* all populations on all processes (processors) is found.  Since the    */
    /* performance measures, etc., have NOT yet been updated in the .stt     */
    /* file, user should now output local_bestfit.xxxxx, etc., not yet */
    /* all_pops_bestfit.xxxx, which will be updated AFTER this print is done.*/
    /* Use of this callback is OPTIONAL. */

}


void
app_print_strings()
/* Optional application-dependent string printer, called by report() just */
/* after it has finished printing the chromosomes in "standard" binary format.*/
{
}

void
app_conv_rept()
/* Application-dependent report, called by reptconv() (in report.c) */
{

}

BOOL
app_user_does_migration(IndArray, filename, number_to_migrate, want_best
				, migration_crowding_amount, ChromSize
				, migration_incest_reduction, BestInd
				, migrant_holder)
/* In this routine, if the user has set the global variable (in app_init)   */
/* call_app_user_does_migration, and user is NOT using inversion, then the  */
/* user HERE is allowed to perform migration among subpopulations in any    */
/* manner the user chooses.  ChromSize is the number of unsigneds in the    */
/* chromosome of the DONOR population, and global chromsize is the number of*/
/* unsigneds in the chromosome of the RECEIVING population (in IndArray).   */
/* IndArray is the receiving population (the global variable is oldpop),    */
/* and transfer occurs at the start of a cycle); the index of the best      */
/* individual in oldpop (in internal form, without one added) is in BestInd.*/
/* For guidance in performing the required operations, see routine          */
/* GetMigrants() in file checkrd.c, where the system-provided migration     */
/* capabilities (migration_incest_reduction, migration crowding, inversion, */
/* etc., are all done.                                                      */
struct individual *IndArray;
char   *filename;
int    number_to_migrate;
BOOL   want_best;
int    migration_crowding_amount, ChromSize, migration_incest_reduction;
int    BestInd;
struct individual * migrant_holder;
{
    return TRUE;
}


BOOL 
app_transform_chrom_to_std(mapped_saved_chromosome, chrom_for_objfunc)
/* If the user is using different representations for different subpops,    */
/* but is NOT using the system-provided inversion code, here is where the   */
/* user must supply the code to transform a migrant individual's chromosome */
/* (in mapped_saved_chromosome) from its "mapped" representation in the */
/* subpopulation to the STANDARD representation, which is the one the user */
/* uses in defining the objective (or fitness) function, objfunc.  The */
/* "standard rep" chromosome must be returned in chrom_for_objfunc, which is */
/* a local name for critter->chrom, the chromosome passed to objfunc().  */ 
/* The user can use any of the global variables, or local variables defined */
/* at the top of this file, to determine which supopulation this is (popno), */
/* and what characteristics THIS representation may have (local variables, */
/* utility fields, etc.) */

unsigned *mapped_saved_chromosome;
unsigned *chrom_for_objfunc;
{
    int ii;

    /* The commented-out code below does the "identity" transformation on the */
    /* chromosome -- i.e., copies one to the other. If the user is doing some */
    /* transformation OTHER than the system-provided inversion, user */
    /* should change the first loop below to perform the actual */
    /* transformation process.  Finally, user must make the function return */
    /* TRUE if successful. */

    /* NOTE:  Next loop must be modified if subpops have different chromosome */
    /* representations -- user must do a transformation, not a copy.  */
/*  for (ii=0;ii<chromsize;ii++) {
	chrom_for_objfunc[ii] = mapped_saved_chromosome[ii];
    }
    return TRUE;
*/
    return FALSE;
}


BOOL
app_transform_migrant(migrant_holder, DonorChromSize, newguy, filename)
/* If the user is using different representations for different subpops,    */
/* but is NOT using the system-provided inversion code, here is where the   */
/* user must supply the code to transform a migrant individual (in a struct */
/* individual pointed to by migrant_holder) from its donor pop rep to its   */
/* popno (a global, the current "receiving" population) representation,     */
/* returning the transformed guy in newguy, which has a chromosome length of*/
/* RecipientChromSize unsigneds.   If user is using inversion, not user's   */
/* "own" transformation, then this routine is not called, so may be left in */
/* its "commented out" form, which is the the IDENTITY transformation (i.e.,*/
/* copy the individual unchanged from migrant_holder to newguy.             */

/* A pointer to the individual to migrate, as read from donorpop, and its   */
/* DonorChromSize, are provided, as is the filename of the recipient pop.   */
/* After transforming the individual, the user should place it into new_guy */
/* and return TRUE if successful.            */
struct individual *migrant_holder, *newguy;
int    DonorChromSize;
char   *filename;
{
    int ii;
/* The commented-out code below does the "identity" transformation.  If user  */
/* is doing some transformation OTHER than the system-provided inversion, user*/
/* should change the first loop below to perform the actual transformation    */
/* process.  Finally, make the function return TRUE if successful. */

/* NOTE:  limit chromsize in next loop below must be revised if subpops have */
/* different chromosome sizes... user must do a transformation, not a copy.  */
/*  for (ii=0;ii<chromsize;ii++) {
	newguy->chrom[ii] = migrant_holder->chrom[ii];
    }
    newguy->parent[0] = migrant_holder->parent[0];
    newguy->parent[1] = migrant_holder->parent[1];
    newguy->xsite[0] = migrant_holder->xsite[0];
    newguy->xsite[1] = migrant_holder->xsite[1];
    newguy->fitness = migrant_holder->fitness;
    newguy->init_fitness = migrant_holder->init_fitness;
    newguy->neval = migrant_holder->neval;
    sys_util_copy_utility(newguy->utility, migrant_holder->utility);

    return TRUE;
*/
    return FALSE;
}


void
app_generate()
/* Opportunity to do any desired (application-dependent) operations at the */
/* end of each generation of each subpopulation. */
{

}


void
app_stats(pop)
/* Application-dependent statistics calculations called by statistics() at  */
/* the end of each generation.                                              */
struct individual *pop;
{
    /* See approyrd.c for an example. */
}


void
app_before_inversion()
/* Allow user to perform any desired application-specific actions before */
/* calling the inversion routine.  Only called when an inversion WILL be */
/* performed on the current subpopulation at the current generation. */
{
}


BOOL
app_write_ckp_hdr(fp)
FILE   *fp;

/* Application-dependent callback routine user may use for writing      */
/* any needed app-specific variables into a checkpoint file.            */
{
    if (fwrite(&field_size, sizeof(int), 1, fp) != 1)
	return FALSE;
    if (fwrite(&vec_size, sizeof(int), 1, fp) != 1)
	return FALSE;
    return TRUE;
}

BOOL
app_read_ckp_hdr(fp)
FILE   *fp;
{
    if (fread(&field_size, sizeof(int), 1, fp) != 1)
	return FALSE;
    if (fread(&vec_size, sizeof(int), 1, fp) != 1)
	return FALSE;
    return TRUE;
}


int
GetUtilitySize()
{
/* Application-dependent callback routine user must use to tell program */
/* the size of the utility field to malloc for each chromosome.  Used */
/* to compute the amount of space needed per individual in the checkpoint */
/* files.  */
    return vec_size * sizeof(int);
}


int
GetMaxUtilitySize()
/* User must here return 0 (utility fields not used) or GetUtilitySize */ 
/* (if different_reps are NOT used), or the size (bytes) of the LARGEST */
/* utility field used by ANY subpopulation, if different_reps includes use */
/* of utility fields of different sizes. */
{
  int nbytes;

  nbytes = GetUtilitySize();
  if (nbytes) {
    if (different_reps)
      /* Here, return size of largest utility field used by any rep. */
     /* nbytes = xxxxxx; */
      return nbytes;
    else
      return nbytes;  /* same as GetUtilitySize if !different_reps */
  }
  return nbytes;
}


