/*----------------------------------------------------------------------------*/
/*checkrd.c  -- routines for reading restart files with the suffix .ind,      */
/*               containing individuals (chromosomes).  Also includes         */
/* the key MIGRATION routines which move individuals among subpopulations.    */
/* Includes functions:                                                        */
/*       read_checkpointfile(), GetMigrants(), GetFitnessBiasedMigrant        */
/*       transform_inversion_pattern(), transform_inversion_pattern_uneq_flds,*/
/*       InitCPRead(), EndCPRead(), NthIndRead(), and FreeUpMigrantHolder()   */
/*----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include "external.h"

#define DEBUG_INVERSION 0
#define RETRIES 10 

/* Defines below are positions in .ind files of individuals. */
#define SEEK_SET                0
#define SEEK_CUR                1 /* as in stdio.h */
#define MVerPos                 0
#define MIndCntPos              (MVerPos + sizeof(int))
#define MIndSizePos             (MIndCntPos + sizeof(int))
#define MChromSizePos           (MIndSizePos + sizeof(int))
#define MBestNowPos             (MChromSizePos+sizeof(int))

static struct individual migrant_holder;
static FILE *fpread;
static int IndCnt;
static int IndSize;
static int ChromSize;
static int BestInd;
static int buf_pos = 0;
static unsigned long mask, indexnum;
static unsigned long MInd0thPos;
static BOOL no_read_intrpt;

#ifdef PROTOTYPES_ACCEPTED
static BOOL NthIndRead(struct individual*, int, BOOL(*)(int*, FILE*));
static BOOL InitCPRead(char *fname, int ownfile, int retries);
static BOOL EndCPRead(char *fname, int ownfile, int retries, BOOL*);
#else
static BOOL NthIndRead();
static BOOL InitCPRead();
static BOOL EndCPRead();
#endif

void 
read_checkpointfile(filename, numberwanted, Must_Recalculate_Bestnow)
     char   *filename;
     int     numberwanted;
     BOOL    Must_Recalculate_Bestnow;
     
/* Opens and reads the .ind (individuals) checkpointfile for restoring a */
/* population, and stores the values read in global data structures */
/* (including oldpop[]).    */
{
  int     j, check;
  double  BestFitnessThisPop = 0.0;

  check = 0;
  do {
    if (InitCPRead(filename, 1, RETRIES) != TRUE) {
      fprintf(stderr, "\nError:InitCPRead failed to restore %s.\n", filename);
      exit (-77);
    }
    /* Note:  InitCPRead read a value for BestInd, a local var, index of best */
    /* guy in file being read. */

    for (j = 0; j < numberwanted; j++) /* NOTE:may not read all individuals */
      if (NthIndRead(&(oldpop[j]), j, sys_util_read_utility) != TRUE) {
	EndCPRead(filename, 1, RETRIES, &no_read_intrpt);
	check++;
	break;
      }
      if (5 <= check) {
	fprintf(stderr,"\nError: %dth individual read for restore failed.\n", 
		j);
	exit (-78);
      }
  } while (j < numberwanted);

  /* Now, if elitism requires that best guy be preserved in a population    */
  /* which is smaller than IndCnt, number of individuals in file being read,*/
  /* read the actual best one here, replace a random guy, and record its    */
  /* index as the new bestnow (note:  BestInd is the index of best indiv in */
  /* the .ind file being read, gotten from the top of the .ind file, and is */
  /* a local variable.  It will be used to get the best guy; then bestnow,  */
  /* the global, will be set to the index of where it appears in oldpop.    */
  if (Must_Recalculate_Bestnow) 
    if (elitism) {
      j = rnd(0, popsize-1);
      if (NthIndRead(&(oldpop[j]), BestInd, sys_util_read_utility) != TRUE) {
	fprintf(stderr, "\nError: Nth individual reading failed at BestInd:");
	fprintf(stderr, " %d.\n", BestInd);
      }
      bestnow = j;
    }
    else {
      /* No elitism, so just find current best, set bestnow to its index*/
      for (j = 0; j < popsize; j++) {
	if (oldpop[j].fitness > BestFitnessThisPop) {
	  BestFitnessThisPop = oldpop[j].fitness;
	  bestnow = j;
	}
      }
    }

  if (EndCPRead(filename, 1, RETRIES, &no_read_intrpt) != TRUE)
    fprintf(stderr, "\nError:EndCPRead failed.\n");
}


/* Reads individuals from a .ind file for migration to neighboring         */
/* subpops.  If call_app_user_does_migration is TRUE (it's a user-settable */
/* global variable which defaults to FALSE), then this routine merely      */
/* calls app_user_does_migration and returns.  Otherwise, reads best (if   */
/* want_best is TRUE) and number_to_migrate-1 others.  If                  */
/* migration_crowding_amount is non-zero, uses DeJong-style crowding (with */
/* the migration crowding amount specified) to select whom to replace in   */
/* receiving subpopulation.   If migration_incest_reduction is non-zero,   */
/* checks that number of randomly selected possible immigrants against the */
/* receiving population's BEST individual, and chooses the one with the    */
/* greatest Hamming distance from it.  In all cases, if elitism is TRUE,   */
/* assures that it does not replace the last copy of the current best      */
/* individual.                                                             */
BOOL
GetMigrants(IndArray, donorpop, filename, number_to_migrate, want_best,
	   migration_incest_reduction, migration_crowding_amount)
     struct individual *IndArray;
     int    donorpop;
     char   *filename;
     int    number_to_migrate, migration_crowding_amount;
     int    migration_incest_reduction;
     BOOL   want_best;
{
  int     i, j, k, m, ii, jj, kk, mm, iii, tmp;
  int     distance, min_distance, max_distance;
  int ownfile;
  double  bestof3fitnesses;
  void    FreeUpMigrantHolder(); 
  BOOL    GetFitnessBiasedMigrant();

  if (startpopnum <= donorpop && donorpop <= finishpopnum)
    /* Donor subpop is run by this process */
    ownfile = 1;
  else
    ownfile = 0;

  if (InitCPRead(filename, ownfile, RETRIES) != TRUE) {
    fprintf(stderr, "  Immigration from %s failed.", filename);
    if (fpread)
      EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt);
    return FALSE;
  }

  /* (Now we have set ChromSize, etc., for donor population) */
  if (call_app_user_does_migration) {
    /* Note: BestInd in call below is local, tells user index of best  */
    /* individual of donor population, in case user wants that.        */
    /* ChromSize is size of chromosome of donor population -- may not  */
    /* match that of recipient population, if different_reps flag is on */
    /* The chromosome below, migrant_holder.chrom, is a newly-created   */
    /* and malloc'd chromosome in struct individual migrant_holder for  */
    /* holding the individual the user will read from the donor         */
    /* population (using ITS ChromSize).  Ultimately, user must place   */
    /* the immigrant into some position in IndArray, which corresponds  */
    /* to the oldpop of the recipient population.                       */
    if ((migrant_holder.chrom = (unsigned *)
	 malloc(ChromSize * sizeof(unsigned))) == NULL)
      nomemory("migrant_holder.chrom");
    sys_util_malloc_utility_field(&(migrant_holder.utility));

    if (app_user_does_migration(IndArray, filename, number_to_migrate
				, want_best, migration_crowding_amount
				, migration_incest_reduction, ChromSize
				, BestInd, &migrant_holder) == TRUE) {
      FreeUpMigrantHolder();
      EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt);
      /* Since have changed the population, must recalculate statistics -- */
      /* some of {best, worst, mean, and scaled .fitness} values will have */
      /* changed.*/
      statistics(oldpop);
      return TRUE;
    }
    else {
      FreeUpMigrantHolder();
      EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt);
      fprintf(stderr, "\napp_user_does_migration() failed.");
      fflush(stderr);
      return FALSE;
    }
  }

  /* Okay, below WE are doing the migration, not the user. */
  for (k = 0; k < number_to_migrate; k++) {
    /* Malloc chromosome in migrant_holder for the guy to read from donor */
    /* population (use ITS ChromSize).                                    */
    /* Note: if different_reps TRUE, disallow migration_incest_reduction. */
    if ((migrant_holder.chrom = (unsigned *)
	 malloc(ChromSize * sizeof(unsigned))) == NULL)
      nomemory("migrant_holder.chrom");
    sys_util_malloc_utility_field(&(migrant_holder.utility));
    if (want_best) {
      i = BestInd;
      /* Only force the first index. */
      want_best = FALSE;
    }
    else if (migration_incest_reduction == 0 || different_reps) {
      if (!GetFitnessBiasedMigrant(&i)) {
	if (fpread)
	  EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt);
	fprintf(stderr, "\nGetFitnessBiasedNMigrant() failed.");
	return FALSE;
      }
    }
    else {
      /* Using incest_reduction to select individual to migrate */
      /* (Not allowed if using different reps in different subpops) */
      /* For each of the migration_incest_reduction individuals to be Hamming */
      /* compared with the best of the receiving subpopulation, pick 3 */
      /* candidates and use the best of the 3 for the Hamming comparison. */
      /* In this way, the candidates for the Hamming comparison are */
      /* pre-selected according to fitness. */
      max_distance = 0;
      for (m = 0; m < migration_incest_reduction; m++) {
	if (!GetFitnessBiasedMigrant(&ii)) {
	  if (fpread)
	    EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt);
	  fprintf(stderr, "\nGetFitnessBiasedNMigrant() failed.");
	  return FALSE;
	}
	if ((distance = hamming(&(migrant_holder), &(IndArray[bestnow])))
	    > max_distance) {
	  max_distance = distance;
	  i = ii;
	}
      }
    }

    /* Now know will migrate individual i, go get it into migrant_holder */
    if (NthIndRead(&(migrant_holder), i, sys_util_read_utility) == FALSE) { 
      int ok = 0;
      no_read_intrpt = TRUE;
      EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt);

      while (!ok && no_read_intrpt == FALSE) {
	no_read_intrpt = TRUE;
	fpread = NULL;
	if (InitCPRead(filename, ownfile, RETRIES) == TRUE) 
	  if (NthIndRead(&(migrant_holder), i, sys_util_read_utility) == TRUE) 
	    ok = 1;

	if (!ok && fpread) 
	  EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt);
      }

      if (!ok) {
	fprintf(stderr, "\nFailed to read %dth individual in GetMigrants()", i);
	fprintf(outfp, "\nFailed to read %dth individual in GetMigrants()", i);
	FreeUpMigrantHolder();
	return FALSE;
      }
    }

    /* Print migrant's chromosome, if user is asking for full output */
    if (printstrings && quiet == 0 && !using_inversion) {
      fprintf(outfp, "\n Migrant sent from subpop %d is:\n", donorpop);
      writechrom(migrant_holder.chrom);
      if (alpha_size != 2)
        writefields(migrant_holder.chrom);
    }

    if (different_reps) {
      if (using_inversion) {
	/* Debug printing */
	if (DEBUG_INVERSION || (printstrings && quiet == 0)) {
	  fprintf(outfp, "\n Migration Event: In donor subpop (%d)", donorpop);
          fprintf(outfp, " (local) rep, migrant is:\n");
	  for (i=0;i<numfields;i++) {
            if (alpha_size <2)
              tmp = getfield_uneq_flds(i,numfields, migrant_holder.chrom
                                    , donor_pop_field_ends);
            else
	      tmp = getfield(i,numfields, migrant_holder.chrom);

	    fprintf(outfp, "%d ", tmp);
	  }
	  fprintf(outfp, "\n");
	}
	
	/* using_inversion; copy old chrom to kidholder[0] */
	for (iii=0;iii<chromsize;iii++) {
	  kidholder[0].chrom[iii] = migrant_holder.chrom[iii];
	}
	/* Now use inversion map to transform it to new rep.       */
        if (alpha_size >= 2) {
	  if (!transform_inversion_pattern(kidholder[0].chrom
					   , this_pop_map_to_std
					   , donor_pop_map_to_std
					   , migrant_holder.chrom)) {
	    fprintf(stderr, "\n*** transformation of chrom failed.\n");
	    FreeUpMigrantHolder();
	    exit(-1);
	  }
        }
        else {
	  if (!transform_inversion_pattern_uneq_flds(kidholder[0].chrom
					   , this_pop_map_to_std
					   , donor_pop_map_to_std
					   , migrant_holder.chrom
                                           , donor_pop_field_ends)) {
	    fprintf(stderr, "\n*** transformation of chrom failed.\n");
	    FreeUpMigrantHolder();
	    exit(-1);
	  }
        }
      }
      else {                    /* Using user-performed transformation */
	/* First, transform OLD migrant_holder chrom into kidholder[0]*/
	/* chrom, which is in the NEW representation. */
	/* (User must write the code to do this change from donor */
	/* pop's representation to recipient pop's representation. */

	app_transform_migrant(&migrant_holder, ChromSize, 
			      &(kidholder[0]), filename);
	/* Transformed migrant is NOW in kidholder[0] -- want to move */
	/* it to migrant_holder    (with new rep chromosome).         */
	FreeUpMigrantHolder();
	/* Make new version of migrant_holder, to hold transformed */
	/* chromosome from  kidholder[0].                          */
	if ((migrant_holder.chrom = (unsigned *)
	     malloc(chromsize * sizeof(unsigned))) == NULL)
	  nomemory("migrant_holder.chrom");
	sys_util_malloc_utility_field(&(migrant_holder.utility));

	/* Now copy the user-transformed chromosome, in new rep,   */
	/* back to the NEW REP migrant_holder chromosome */
	for (iii = 0; iii < chromsize; iii++) 
	  migrant_holder.chrom[iii] = kidholder[0].chrom[iii];

	migrant_holder.parent[0] = kidholder[0].parent[0];
	migrant_holder.parent[1] = kidholder[0].parent[1];
	migrant_holder.xsite[0] = kidholder[0].xsite[0];
	migrant_holder.xsite[1] = kidholder[0].xsite[1];
	migrant_holder.fitness = kidholder[0].fitness;
	migrant_holder.init_fitness = kidholder[0].init_fitness;
	migrant_holder.neval = kidholder[0].neval;
	sys_util_copy_utility(migrant_holder.utility, kidholder[0].utility);
      }
      /* Now the guy to migrate is in migrant_holder, just as if he */
      /* had originated from a pop with same rep as recipient pop.  */
    }


    /* Now, must decide where to put the immigrant individual. */
    /* (Whether is different_rep or NOT.) */
    /* Choose individual [j] to replace in IndArray */
    /* If migration_crowding_amount non-zero, repeat the loop to choose */
    /* among several possible candidates. */
    /* NOTE:  although this process prevents replacing the last copy of best */
    /* guy when elitism is on, it does NOT prevent one migrant from replacing */
    /* another just introduced... that happens randomly.  New migrants are */
    /* not protected any more than oldpop "survivors" were.  */
    min_distance = lchrom + 1;
    kk = 0;

    do {
      /* Pick m, a possible candidate for j, the one to replace. */
      m = rnd(0, popsize - 1);
      /* If elitism is on, make sure don't replace only copy of best */
      if (elitism) {
	if (IndArray[bestnow].init_fitness ==IndArray[m].init_fitness) {
	  mm = m;
	  /* To preserve fairness, first see if this is UNIQUE best */
	  jj=0;
	  while (jj<popsize) {
	    if (jj != m && IndArray[bestnow].init_fitness
		== IndArray[jj].init_fitness) {
	      /* Not unique, okay to use original m picked, */
	      /* terminate while clause. */
	      break;
	    }
	    else jj++;
	  }
	  if(jj == popsize) {
	    /* No other bests found, need a different m */
	    while (m == mm)
	      m = rnd(0, popsize - 1);
	  }
	}
      }
      /* Okay, m is an okay index to replace, whether elitism on or not.*/
      /* Pick the one closest to the receiving best, to replace. */
      /* (Note:  always true the first time through, so j always set */
      if ((distance = hamming(&(IndArray[m]), &(migrant_holder)))
	  < min_distance) {
	j = m;
	min_distance = distance;
      }
      kk++;

    } while (kk < migration_crowding_amount);

    /* Okay, j is the one to replace, so do it (differently if         */
    /* using different_reps than if not). */
    if (!different_reps) {
      if (NthIndRead(&(IndArray[j]), i, sys_util_read_utility) != TRUE) {
	fprintf(stderr, "\nError: Reading failed at order:");
	fprintf(stderr, " %d in RandomRead.\n", BestInd);
	EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt);
	FreeUpMigrantHolder();
	return FALSE;
      }
      /* Here, the migration was successful; free up the migrant_holder */
      FreeUpMigrantHolder();
    }
    else {                      /* Different representations in use */
      for (iii=0;iii<chromsize;iii++) {
	IndArray[j].chrom[iii] = migrant_holder.chrom[iii];
      }
      IndArray[j].parent[0] = migrant_holder.parent[0];
      IndArray[j].parent[1] = migrant_holder.parent[1];
      IndArray[j].xsite[0] = migrant_holder.xsite[0];
      IndArray[j].xsite[1] = migrant_holder.xsite[1];
      IndArray[j].fitness = migrant_holder.fitness;
      IndArray[j].init_fitness = migrant_holder.init_fitness;
      IndArray[j].neval = migrant_holder.neval;
      sys_util_copy_utility(IndArray[j].utility, migrant_holder.utility);
      FreeUpMigrantHolder();
    }
    /* Debug Print Here */
    if (DEBUG_INVERSION) {
      fprintf(outfp, "\n In receiving oldpop (new local) rep, migrant ");
      fprintf(outfp, "(printed from newpop) is:\n");
      for (i=0;i<numfields;i++) {
	tmp = getfield(i,numfields, oldpop[j].chrom);
	fprintf(outfp, "%d ", tmp);
      }
      fprintf(outfp, "\n");
    }

  }

  if (EndCPRead(filename, ownfile, RETRIES, &no_read_intrpt) != TRUE) 
    if (no_read_intrpt == TRUE) {
      fprintf(stderr, "\nError:EndCPRead failed.\n");
      return FALSE;
    }
  /* Since have changed the population, must recalculate statistics -- some */
  /* of the {best, worst, mean, and scaled .fitness} values will have changed.*/
  statistics(oldpop);
  return TRUE;
}


void
FreeUpMigrantHolder()
/* Routine simply frees up the migrant_holder mallocs */
{
  free(migrant_holder.chrom);
  migrant_holder.chrom = NULL;
  if (migrant_holder.utility != NULL) {
    free(migrant_holder.utility);
    migrant_holder.utility = NULL;
  }
  return;
}


BOOL
GetFitnessBiasedMigrant(jj)
int *jj;
/* Pick 3 individuals from the neighboring subpop; use tournament selection */
/* to keep best one (regardless of what selection is in use for reproduction) */
{
  int   j, ii;
  double bestof3fitnesses; 

  bestof3fitnesses = -1.;
  for (j=0;j<3;j++) {
    ii = rnd(0, IndCnt - 1);
    /* Below, use migrant_holder as temp storage for individual */
    if (NthIndRead(&(migrant_holder), ii
		   , sys_util_read_utility) != TRUE) {
      fprintf(stderr, "***OOPS, Migration Error in incest_reduction.");
      fprintf(outfp, "***OOPS, Migration Error in incest_reduction.");
      FreeUpMigrantHolder();
      return FALSE;
    }
    if (migrant_holder.init_fitness > bestof3fitnesses) { 
      bestof3fitnesses = migrant_holder.init_fitness;
      *jj = ii;
    }
  }
  return TRUE;
}


/* First transforms donor chrom into "std" rep used by objective function, */
/* then transforms THAT (inversely from usual transform) to rep of the     */
/* receiving population.                                                   */
BOOL
transform_inversion_pattern(chrom_in_donor_pattern, this_pop_map_to_std,
			    donor_pop_map_to_std, chrom_in_new_pattern)
     unsigned *chrom_in_donor_pattern, *chrom_in_new_pattern;
     int *this_pop_map_to_std, *donor_pop_map_to_std;
{
  int i, tmp;

  /* Debug printing */
  if (DEBUG_INVERSION) {
    fprintf(outfp, "\n donor_pop_map_to_std is:\n");
    for (i=0;i<numfields;i++) {
      fprintf(outfp, "%d ", donor_pop_map_to_std[i]);
    }
    fprintf(outfp, "\n this_pop_map_to_std is:\n");
    for (i=0;i<numfields;i++) {
      fprintf(outfp, "%d ", this_pop_map_to_std[i]);
    }
    fprintf(outfp, "\n Orig (local rep.) chrom is:\n");
    for (i=0;i<numfields;i++) {
      tmp = getfield(i,numfields,chrom_in_donor_pattern);
      fprintf(outfp, "%d ", tmp);
    }
  }
  if (!map_chromosome_to_std(chrom_in_donor_pattern
                             , donor_pop_map_to_std
			     , chrom_in_new_pattern)) {
    fprintf(stderr, "\n*** Remapping migrating chrom failed.\n");
    return FALSE;
  }
  /* Debug printing */
  if (DEBUG_INVERSION || (printstrings && quiet == 0)) {
    fprintf(outfp, "\n In STANDARD rep, chrom is:\n");
    for (i=0;i<numfields;i++) {
      tmp = getfield(i,numfields,chrom_in_new_pattern);
      fprintf(outfp, "%d ", tmp);
    }
  }
  /* Now copy it back, (saves a buffer). */
  for (i=0;i<chromsize; i++)
    chrom_in_donor_pattern[i] = chrom_in_new_pattern[i];
  /* Debug printing */
  if (DEBUG_INVERSION) {
    fprintf(outfp, "\n Again, STANDARD chrom is:\n");
    for (i=0;i<numfields;i++) {
      tmp = getfield(i,numfields,chrom_in_donor_pattern);
      fprintf(outfp, "%d ", tmp);
    }
  }
  /* Now use inverse of map_chromosome_to_std with receiving pop map */
  for (i=0;i<numfields;i++) {
    putfield(i
	     , numfields
	     , getfield(this_pop_map_to_std[i]
			, numfields
			, chrom_in_donor_pattern)
	     , chrom_in_new_pattern);
  }
  /* Debug printing */
  if (DEBUG_INVERSION || (printstrings && quiet == 0)) {
    fprintf(outfp, "\n In receiving subpop %d (local) rep, chrom is:\n",popno);
    for (i=0;i<numfields;i++) {
      tmp = getfield(i,numfields,chrom_in_new_pattern);
      fprintf(outfp, "%d ", tmp);
    }
    /* For Debug, map it back to std and print, to check process */
    if (!map_chromosome_to_std(chrom_in_new_pattern, this_pop_map_to_std
			       , chrom_in_donor_pattern)) {
      fprintf(stderr, "\n*** Remapping migrating chrom failed.\n");
      return FALSE;
    }
    /* Debug printing */
    if (DEBUG_INVERSION) {
      fprintf(outfp, "\n In STANDARD rep, chrom is:\n");
      for (i=0;i<numfields;i++) {
        tmp = getfield(i,numfields, chrom_in_donor_pattern);
        fprintf(outfp, "%d ", tmp);
      }
      fprintf(outfp, "\n");
    }
  }
  return TRUE;
}


/* For cases when field lengths are NOT the same, */
/* first transforms donor chrom into "std" rep used by objective function, */
/* then transforms THAT (inversely from usual transform) to rep of the     */
/* receiving population.                                                   */
BOOL
transform_inversion_pattern_uneq_flds(chrom_in_donor_pattern
                            , this_pop_map_to_std
                            , donor_pop_map_to_std, chrom_in_new_pattern
                            , donor_pop_ends_of_fields)
     unsigned *chrom_in_donor_pattern, *chrom_in_new_pattern;
     int *this_pop_map_to_std, *donor_pop_map_to_std;
     int *donor_pop_ends_of_fields;
{
  int i, tmp;

  /* Debug printing */
  if (DEBUG_INVERSION) {
    fprintf(outfp, "\n donor_pop_map_to_std is:\n");
    for (i=0;i<numfields;i++) {
      fprintf(outfp, "%d ", donor_pop_map_to_std[i]);
    }
    fprintf(outfp, "\n donor_pop_field_ends is:\n");
    for (i=0;i<numfields;i++) {
      fprintf(outfp, "%d ", donor_pop_ends_of_fields[i]);
    }
    fprintf(outfp, "\n this_pop_map_to_std is:\n");
    for (i=0;i<numfields;i++) {
      fprintf(outfp, "%d ", this_pop_map_to_std[i]);
    }
    fprintf(outfp, "\n this pop's field_ends array is:\n");
    for (i=0;i<numfields;i++) {
      fprintf(outfp, "%d ", field_ends[i]);
    }
    fprintf(outfp, "\n Orig (local rep.) chrom is:\n");
    for (i=0;i<numfields;i++) {
      tmp = getfield_uneq_flds(i,numfields,chrom_in_donor_pattern
                                  , donor_pop_ends_of_fields);
      fprintf(outfp, "%d ", tmp);
    }
  }
  if (!map_chromosome_to_std_uneq_flds(chrom_in_donor_pattern
                             , donor_pop_map_to_std
			     , chrom_in_new_pattern
                             , donor_pop_ends_of_fields)) {
    fprintf(stderr, "\n*** Remapping migrating chrom failed.\n");
    return FALSE;
  }
  /* Debug printing */
  if (DEBUG_INVERSION || (printstrings && quiet == 0)) {
    fprintf(outfp, "\n In STANDARD rep, chrom is:\n");
    for (i=0;i<numfields;i++) {
      tmp = getfield_uneq_flds(i,numfields,chrom_in_new_pattern
                                  , std_field_ends);
      fprintf(outfp, "%d ", tmp);
    }
  }
  /* Now copy it back, (saves a buffer). */
  for (i=0;i<chromsize; i++)
    chrom_in_donor_pattern[i] = chrom_in_new_pattern[i];
  /* Debug printing */
  if (DEBUG_INVERSION) {
    fprintf(outfp, "\n Again, STANDARD chrom is:\n");
    for (i=0;i<numfields;i++) {
      tmp = getfield_uneq_flds(i,numfields,chrom_in_donor_pattern
                                  , std_field_ends);
      fprintf(outfp, "%d ", tmp);
    }
  }
  /* Now use inverse of map_chromosome_to_std with receiving pop map */
  for (i=0;i<numfields;i++) {
    putfield(i
	     , numfields
	     , getfield_uneq_flds(this_pop_map_to_std[i]
			, numfields
			, chrom_in_donor_pattern
                        , std_field_ends)
	     , chrom_in_new_pattern);
  }
  /* Debug printing */
  if (DEBUG_INVERSION || (printstrings && quiet == 0)) {
    fprintf(outfp, "\n In receiving subpop %d (local) rep, chrom is:\n",popno);
    for (i=0;i<numfields;i++) {
      tmp = getfield(i,numfields,chrom_in_new_pattern);
      fprintf(outfp, "%d ", tmp);
    }
    /* For Debug, map it back to std and print, to check process */
    if (!map_chromosome_to_std(chrom_in_new_pattern, this_pop_map_to_std
			       , chrom_in_donor_pattern)) {
      fprintf(stderr, "\n*** Remapping migrating chrom failed.\n");
      return FALSE;
    }
    /* Debug printing */
    if (DEBUG_INVERSION) {
      fprintf(outfp, "\n In STANDARD rep, chrom is:\n");
      for (i=0;i<numfields;i++) {
        tmp = getfield_uneq_flds(i,numfields, chrom_in_donor_pattern
                                    , std_field_ends);
        fprintf(outfp, "%d ", tmp);
      }
      fprintf(outfp, "\n");
    }
  }
  return TRUE;
}


static BOOL
InitCPRead(fname, ownfile, retries)
     char   *fname;
     int ownfile, retries;
{
  int version;

  if (!(fpread = start_read(fname, ownfile, retries, &mask, &indexnum, &buf_pos)))
    {
      fprintf(stderr, "\nUnable to open file %s.", fname);
      return FALSE;
    }

  if ((fread(&version, sizeof(int), 1, fpread)))
    if (fread(&IndCnt, sizeof(int), 1, fpread))
      if (fread(&IndSize, sizeof(int), 1, fpread))
	if (fread(&ChromSize, sizeof(int), 1, fpread))
	  if (fread(&BestInd, sizeof(int), 1, fpread)) {
	    MInd0thPos = ftell(fpread);
	    return TRUE;
	  }

  MInd0thPos = 0;
  if (fpread)
    fclose(fpread);
  return FALSE;
}


static BOOL
EndCPRead(fname, ownfile, retries, read_cmplt)
char *fname;
int ownfile, retries;
BOOL *read_cmplt;
{
  if ((end_read(fname, fpread, ownfile, retries, 
		mask, indexnum, buf_pos, read_cmplt))) {
    fpread = NULL;
    return TRUE;
  }
  
  fpread = NULL;
  return FALSE;
}


static BOOL
NthIndRead(IndBuf, order, CBfunc)
     struct individual *IndBuf;
     int     order;
     BOOL(*CBfunc) ( /* int *,FILE * */ );
{
  long thisIndPos = MInd0thPos + order * IndSize;
  int check = 0;

  if (!fpread || (fseek(fpread, thisIndPos, SEEK_SET))) {
    if (!fpread)
      fprintf(stderr, "No file pointer in NthIndRead().\n");
    else 
      fprintf(stderr, "Can't set filepointer to %ld in NthIndRead(%d).\n", 
	      thisIndPos, order);
    return FALSE;
  }

  if (fread(IndBuf->chrom, sizeof(unsigned), ChromSize, fpread)) 
    if ((check = fread(&(IndBuf->fitness), sizeof(double), 1, fpread))) 
      if (fread(&(IndBuf->init_fitness), sizeof(double), 1, fpread))
	if (fread(&(IndBuf->neval), sizeof(long), 1, fpread))
	  if (fread(&(IndBuf->xsite[0]), sizeof(int), 1, fpread))
	    if (fread(&(IndBuf->xsite[1]), sizeof(int), 1, fpread))
	      if (fread(&(IndBuf->parent[0]), sizeof(int), 1, fpread))
		if (fread(&(IndBuf->parent[1]), sizeof(int), 1, fpread)) 
		  return (*CBfunc)(IndBuf->utility, fpread);

  return FALSE;
}


BOOL
sys_util_read_utility(UtilityBuf, fp)
     int    *UtilityBuf;
     FILE   *fp;
     
/* Application-dependent callback routine which provides for reading the  */
/* contents of the utility field from a checkpoint file, if utility fields */
/* were written there by sys_util_write_utility() (GetUtilitySize() > 0).  */
{
  unsigned nbytes;

  nbytes = GetUtilitySize();
  if (nbytes)
    if (fread(UtilityBuf, nbytes, 1, fp) != 1)
      return FALSE;
  return TRUE;
}











