/*----------------------------------------------------------------------------*/
/* initsubp.c - gets things set up and initialized for running multiple       */
/*             subpopulations from user input or from a numbered series of    */
/*             checkpoint files.     */
/*             Allows resetting of parameters by user to values different     */
/*             from what were in restart file, or defaulting of all but       */
/*             maxgen to values used when file was run.  In all cases, at     */
/*             EXTEND  time, the default values presented to the user are the */
/*             values from the saved checkpoint file.  Checkpt files ALWAYS   */
/*             written.                                                       */
/*----------------------------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <limits.h>
#include "external.h"

static int numberwantfromfile;  /* no. of indiv's to be read from .ind file */


/* Initialization Coordinator */
/* Manypops ALWAYS performs any EXTEND run from the startpopnum subpopulation. */
/* A NOTE on file name suffixes (extensions):  Manypops writes two different */
/* TYPES of checkpoint file:  "individual" files, containing members of the */
/* population, and "header" files, containing all other state information */
/* about the definition and status of a run.  There is a separately numbered */
/* file of each type for each subpopulation.  However, in order to maintain a */
/* stable configuration, which allows EXTENDS even after execution is inter- */
/* rupted in mid-cycle, Manypops writes into TWO different buffers in each */
/* checkpoint file (two in .ckp and two in .ind files).  It alternates in */
/* their writing */
/* so that, when other subpopulations (in the same process) READ migrants from*/
/* them, the buffer WRITTEN IN THE PREVIOUS CYCLE is read, so that migrants */
/* from subpopulations "ahead" and "behind" a receiving subpopulation are */
/* from the */
/* SAME cycle (namely, the previous one).  */
/* Then, ONLY AFTER an ENTIRE CYCLE is COMPLETED, the indicator for the */
/* buffer to be READ from is updated in ALL of the checkpoint files for that */
/* process's */
/* subpopulations.  Thus, even when a run is interrupted in mid-cycle, the */
/* checkpoint */
/* files always represent a consistent picture of ALL populations at the END */
/* of a */
/* FULL CYCLE.  Then, if an interrupted run is EXTENDED, some calculation is */
/* lost, */
/* but all measures of performance, counts of evaluations, etc., which have */
/* any bearing on the EXTENDED run are correct.  Additional note:  stats   */
/* such as number of evaluations performed in ALL pops, on-line and off-line */
/* performance for ALL populations, and best individual ever found over ALL  */
/* subpops, etc., are stored in a single file, the .stt file, used by ALL the*/
/* processors (one or more) running in this directory.  Its use allows the */
/* subpopulations to run completely   */
/* asynchronously from one another, and statistics printed are always up to  */
/* date AS OF THE TIME the cycle started for the current subpopulation.      */
/* That is, statistics from other pops are read at the beginning of each     */
/* subpop's cycle, and adjusted locally during the cycle for that subpop,    */
/* and then the all_pops data is updated with those changes by READING IT    */
/* again (to catch any updates already made by another population running in */
/* another process), keeping  it LOCKED, and then writing the updated numbers */
/* to it before unlocking it.*/
void
initialize_parallel(askrevise)
     BOOL    *askrevise;
{
  unsigned  nbytes;
  int     indcnt;               /* number of indiv's available in .ind file */
  int     i;                    /* index */
  char    answer = 'y';
  BOOL    Must_Recalculate_Bestnow = FALSE;

  static int firsttimecalled = 1;     /* Another one. */
  static int starting_guy_index; /* index of individual where a type of */
				    /* initialization is to begin */
  static BOOL random_never_initialized = TRUE;

  CombineFileName(checkpthdrfilename, checkptfileprefix, popno, ".ckp");
  CombineFileName(checkptindfilename, checkptfileprefix, popno, ".ind");

  /* If a restart file name was given, the run will be an EXTEND of a */
  /* run already made, or we are in the middle of a run, so prepare to */
  /* open and read restart header file, from which we will get the saved */
  /* params.  Use this "restart" filename only for the first cycle; */
  /* restart later cycles from the new ckpt files being written during */
  /* this EXTENDING of the run (code in mainmany.c copies the */
  /*  checkptfileprefix into restartfileprefix after first cycle is complete).*/
  if (strlen(restartfileprefix)) {
    /* THIS IS THE CODE FOR RESTARTING FROM A SAVED CHECKPOINT FILE, */
    /* WHETHER EXTENDING A PREVIOUS RUN, OR CONTINUING ONE ALREADY IN */
    /* PROGRESS (just restoring a subpop from its saved checkpointfile) */
    CombineFileName(restarthdrfilename, restartfileprefix, popno, ".ckp");
    CombineFileName(restartindfilename, restartfileprefix, popno, ".ind");

    indcnt = ReadCheckPointHeader(restarthdrfilename, app_read_ckp_hdr);
    if (indcnt == -1) {
      fprintf(outfp, "\n *** Sorry, Unable to read the");
      fprintf(outfp, " restart header file (.ckp). ***\n");
      fflush(outfp);
      exit(-1);
    }
    startgen = maxgen;          /* just starting where it left off before, */
				/* with gen number at end of prev. cycle. */
    /* askrevise is always TRUE first cycle starting or restarting run, */
    /* never if cycle > 0 */
    if (*askrevise == TRUE) {
      if (interactive) {
	printf("Change any parameters for subpop %d? (y) > ", popno);
	fflush(stdout);
	ffscanf("changes_to_this_pop", infp, "%c", &answer);
      }
      else if (!(ffscanf("changes_to_this_pop", infp, "%c", &answer))) {
	fprintf(outfp,
		"\n*** End of Input File, assuming 'No More Changes'\n");
	fprintf(stderr,
		"\n*** End of Input File, assuming 'No More Changes'\n");
	*askrevise = FALSE;
      }

      if (answer == 'y' || answer == 'Y')
	/* allow revision of parameters with defaults from file */
	subpop_data(&Must_Recalculate_Bestnow);
    }
    /* If are parameter changes on cmd line, and NOT interactive, then */
    /* cmd line values are more current -- use them */
    if (parameter_changes && !interactive)
      get_pop_overrides();

    /* Continue with initialization */
    initmalloc();

    /* Malloc the utility fields for the two populations, newpop and oldpop */
    sys_util_malloc();

    numberwantfromfile = popsize > indcnt ? indcnt : popsize; /* min */
    read_checkpointfile(restartindfilename, 
			numberwantfromfile, Must_Recalculate_Bestnow);

    /* Give application a chance to do its thing; it must figure  */
    /* out that it is a restart by itself.  */
    app_init();
    /* If reading all popsize individuals, next calls are NOP's. */
    starting_guy_index = numberwantfromfile;
    if (user_supplied_initialization)
      app_user_init_pop(starting_guy_index);
    else if (permproblem)
      initpermpop(starting_guy_index);
    else if (alpha_size != 2)
      init_non_binary(starting_guy_index);
    else if (superuniform_initialization)
      init_superuniform_pop(starting_guy_index);
    else
      init_random_pop(starting_guy_index);

    /* Let user clean up or change or do over the initialization now. */
    app_after_random_init();

    /* Here is where individuals to be migrated in from neighbors */
    /* at the beginning of a cycle are brought in to this subpop, if */
    /* a master file was provided. */
    if (masterfileprefix[0] != '@')
      ReadFromNeighbor(popno);

    statistics(oldpop);
    if (quiet < 3 && cycle == 0)
      initreport_parallel();

    /* Next, check to be sure user has not just turned on inversion. */
    if (pinversion > 0.) {
      using_inversion = 1;
      different_reps = 1;
    }
  }

  else {
    /* THIS IS THE CODE FOR INITIALIZING FROM SCRATCH, NOT FROM A CHECKPOINT 
       FILE. WE WILL ONLY COME HERE IN CYCLE 0, WHEN NOT EXTENDING A RUN */
    numberwantfromfile = 0;
    startgen = 0;
    gen = 0;
    one_pop_cum_sumfitness = 0.;
    one_pop_sum_best_fitness = 0.;
    one_pop_best_fitness_count = 0;
    one_pop_online_denominator = 0;

    /* get basic problem values from input stream */
    initdata();

    /* malloc the populations and various fixed-length structures */
    initmalloc();

    /* Let user malloc the utility fields for the populations */
    sys_util_malloc();
    
    /* initialize application dependent variables */
    app_init();

    if (firsttimecalled) {
      firsttimecalled = 0;
      bestever_variable_malloc();
    }

    /* initialize random number generator */
    if(random_never_initialized) {
      Randomize(&randomseed);
      random_never_initialized = FALSE;
    }

    /* initialize global counters/values */

    nmutation = 0;
    ncross = 0;
    neval = 0;
    bestfit.fitness = -1.e100;
    bestfit.init_fitness = -1.e100;
    bestfit.neval = 0;
    bestfit.generation = 0;
    maxgen = genspercycle;
    windowstart = 0;
    windowend = 0;


    /* initialize the population */
    starting_guy_index = 0;
    if (user_supplied_initialization)
      app_user_init_pop(starting_guy_index);
    else if (permproblem)
      initpermpop(starting_guy_index);
    else if (alpha_size != 2)
      init_non_binary(starting_guy_index);
    else if (superuniform_initialization)
      init_superuniform_pop(starting_guy_index);
    else
      init_random_pop(starting_guy_index);

    /* Let user clean up or change or do over the initialization now. */
    app_after_random_init();

    /* report statistics */
    statistics(oldpop);
    windowend++;
    if (quiet <3)
      initreport_parallel();
  }
}


/* data inquiry and setup at start of a run. */
void
initdata()
{
  BOOL f = FALSE;
  int i, k, prev_end, fldsiz, fldlen, nbytes;
  int *inverse_map_to_std;

  static  BOOL firstcall = TRUE;

  if (firstcall) {
    startgen = 0;
    superuniform_initialization = FALSE;
    genspercycle = 10;
    popsize = -1;
    pcross = -1.0;
    pmutation = -1.0;
    pinversion = 0.0;
    scaling_window = -1;
    sigma_trunc = 0.0;
    scalemult = 0;
    crowding_factor = 0;
    incest_reduction = 0;
    conv_sigma_coeff = 7.0;
    convinterval = 10;
    gen = 0;
  }

  else if (0 < cycle)
    return;

  else if (all_subpops_use_same_parameters) {
    if (parameter_changes && !interactive)
      get_pop_overrides();
  }

  if (interactive) {
    if (all_subpops_use_same_parameters == TRUE)
      fprintf(outfp,"\n   ---*** Data Entry for Subpopulations ");
    else
      fprintf(outfp,"\n   ---*** Data Entry for Subpopulation %d ", popno);
    fprintf(outfp,"***---\n\n");
  }

  if (firstcall || !all_subpops_use_same_parameters) { 
    subpop_data(&f);
    if (parameter_changes && !interactive)
      get_pop_overrides();
  }

  /* Now, regardless of alpha_size, get needed arrays malloc'd */  
  /* If max sizes for bestever variables was not set by user in function */
  /* app_set_options, do something here which should be correct. */
  if (max_numfields_among_all_subpops == -1)
    max_numfields_among_all_subpops = numfields;
  
  malloc_field_sizes_arrays();          

  if (alpha_size >= 2) {
    /* Set up array field_sizes so can use it, even though all same. */
    for (i=0;i<numfields;i++) {
      field_sizes[i] = alpha_size;
    }
    /* Since field_sizes all equal, get ceiling log base 2 of alpha_size for */
    /* fieldlength, use to set field_ends[].  */
    k = 1;
    for (fieldlength = 1; fieldlength < 17; fieldlength++) {
      k *= 2;
      if (k >= alpha_size)
	break;
    }
    field_ends[0] = fieldlength;
    for (i=1;i<numfields;i++) {
      field_ends[i] = field_ends[i-1] + fieldlength;
    }
  }
    
  /* Else, must get length (bits) of each field specified in field_sizes[] */
  else {
    /* Set to illegal values, so can be sure user set them. */
    for (i=0;i<numfields;i++)
      field_sizes[i] = -1;
    app_set_field_sizes();
    /* Now be sure user set all of them to legal values */
    for (i=0;i<numfields;i++)
      if (field_sizes[i] < 2) {
	fprintf(stderr, "*** ERROR.  User set alpha_size < 2, indicating");
	fprintf(stderr, " that field_sizes are\ndifferent from field to ");
	fprintf(stderr, "field.  However, user did NOT specify legal ");
	fprintf(stderr, "field sizes\nfor all %d fields in ", numfields);
	fprintf(stderr, "function app_set_field_sizes() in the ");
	fprintf(stderr, "app-specific\nfile.  Run cannot continue.  ***\n");
	fprintf(outfp, "*** ERROR.  User set alpha_size < 2, indicating");
	fprintf(outfp, " that field_sizes are\ndifferent from field to ");
	fprintf(outfp, "field.  However, user did NOT specify legal ");
	fprintf(outfp, "field sizes\nfor all %d fields in ", numfields);
	fprintf(outfp, "function app_set_field_sizes() in the ");
	fprintf(outfp, "app-specific\nfile.  Run cannot continue.  ***\n");
	exit (-1);
      }
    prev_end = 0;
    for (i=0;i<numfields;i++) {
      fldsiz = field_sizes[i];
      k = 1;
      for (fldlen = 1; fldlen < 17; fldlen++) {
	k *= 2;
	if (k >= fldsiz)
	  break;
      }
      field_ends[i] = prev_end + fldlen;
      prev_end += fldlen;
    }
  }

  /* Malloc below is handled as a special case, because it must NOT */
  /* be done in initmalloc() if restarting from checkptfile (then   */
  /* it MUST be done inside ReadCheckPointHeader).  Therefore, it is  */
  /* done separately here, for the case NOT restarting from a saved */
  /* checkpoint file */
  nbytes = numfields * sizeof(int);
  if ((this_pop_map_to_std = (int *) malloc(nbytes)) == NULL)
    nomemory("this_pop_map_to_std");
    
  /* Even if not using inversion, initialize map to identity. If user */
  /* wants to start with a different map (for any but "standard"  */
  /* population, user can override it in app_init below, for example. */
  initialize_inversion_map(this_pop_map_to_std);
    
  if (firstcall) {                      
    firstcall = FALSE;
    /* Create this info for Standard population, in case later */
    /* is changed by inversion, etc.  Might not be the rep of any subpop */
		      
    if (popno == startpopnum) {                  
      /* Allocate temporary memory for inverse of map_to_std array */
      nbytes = numfields * sizeof(int);
      if ((inverse_map_to_std = (int *) malloc(nbytes)) == NULL)
	nomemory("inverse_map_to_std");
      /* Create inverse map to std using temporary array just malloc'd */
      for (i=0;i<numfields;i++) 
	inverse_map_to_std[this_pop_map_to_std[i]] = i; 

      /* Now, use that to calculate std_field_ends, the field_ends array */
      /* for use with the STANDARD representation */
      for (i=0;i<numfields;i++) {
	std_field_ends[i] = (i==0 ? 0 : std_field_ends[i-1]) 
		    + (field_ends[inverse_map_to_std[i]] 
		      - (i==0 ? 0 : field_ends[inverse_map_to_std[i]-1]));
      }
      free(inverse_map_to_std);
      inverse_map_to_std = NULL;
    }

/* do rest for all subpops */
    lchrom = field_ends[numfields - 1];
    /* define chromosome size in terms of machine bytes, ie  */
    /* length of chromosome in bits (lchrom)/(bits-per-byte) */
    /* chromsize must be known for malloc() of chrom pointer */
    chromsize = (lchrom / UINTSIZE);
    if (lchrom % UINTSIZE)
      chromsize++;
    if (interactive)
      printf("   * Each chromosome will be %d bit(s), %d unsigned(s) long.\n", 
	     lchrom, chromsize);
    if (max_chromsize_among_all_subpops == -1)
      max_chromsize_among_all_subpops = chromsize;

  
  }
}


void
init_random_pop(starting_guy_index)
     /* Initialize a population at random */
     int starting_guy_index;
{
  int     j, k;
  unsigned rnd_max_uint();

  for (j = starting_guy_index; j < popsize; j++) {
    for (k = 0; k < chromsize; k++)
      oldpop[j].chrom[k] = rnd_max_uint();

    Initialize_For_New_Individual(j);
  }
}

void
init_non_binary(starting_guy_index)
     /* Initialize a population at random for non-binary, */
     /* non-permutation-type problem (alpha_size != 2, not nec. a pwr of 2 */
     int starting_guy_index;
{
  int     i, j, start;
  int     randvalue;

  for (j = starting_guy_index; j < popsize; j++) {

    /* write random fields (in range) to their places in chromosome */
    start = 1;
    for (i = 0; i < numfields; i++) {
      randvalue = rnd(0, field_sizes[i] - 1);
      int2ithruj(start, field_ends[i], randvalue, oldpop[j].chrom);
      start = field_ends[i] + 1;
    }
    Initialize_For_New_Individual(j);
  }
}

/* Perform superuniform initialization of a BINARY population, starting at */
/* population index starting_guy_index (lower ones were read from file) */
/* and continuing for the largest power of 2 remaining below popsize.  For */
/* any remaining guys, initialize them at random.  Note:  OPTIONAL */
int starting_guy_index;
void
init_superuniform_pop(starting_guy_index)
{
  int     j, k, n_superuniform_initialize, stopping_guy_index;
  int     num_superuniform_cycles, bits_left_over, startvalue;
  int     cycle_index;
  int     num_patterns_with_bits_left_over, startbit, stopbit;

  if(starting_guy_index == popsize)
    return;
  superuniform_cyclebitwidth = floorlogbase2(popsize - starting_guy_index) ;
  n_superuniform_initialize = (int)powerof2(superuniform_cyclebitwidth) ;
  stopping_guy_index = starting_guy_index + n_superuniform_initialize;
  num_superuniform_cycles = lchrom/superuniform_cyclebitwidth; /* floor */
  bits_left_over = lchrom % superuniform_cyclebitwidth; /* special case  */
  num_patterns_with_bits_left_over = (int)powerof2(bits_left_over);

  startbit = 1;
  stopbit = superuniform_cyclebitwidth;
  for (cycle_index=0; cycle_index< num_superuniform_cycles; cycle_index++) {

    /* Pick a random value in the range to start the sequence at.  */
    startvalue = rnd(0, (int)(powerof2(superuniform_cyclebitwidth) - 1));

    /* Write values for this field to their places on the chromosomes */
    for (j = starting_guy_index; j < stopping_guy_index; j++) {
      k = (j + startvalue) % n_superuniform_initialize;
      int2ithruj(startbit, stopbit, k, oldpop[j].chrom);
    }

    startbit += superuniform_cyclebitwidth;
    stopbit += superuniform_cyclebitwidth;
  }

  /* Now handle the final partial-width field (if any) */
  stopbit = startbit + bits_left_over - 1;
  startvalue = rnd(0, num_patterns_with_bits_left_over - 1);
  for (j = starting_guy_index; j < stopping_guy_index; j++) {
    if(stopbit >= startbit) {
      k = (j + startvalue) % num_patterns_with_bits_left_over;
      int2ithruj(startbit, stopbit, k, oldpop[j].chrom);
    }
    /* and fill in the rest of the fields, while we're at it. */
    Initialize_For_New_Individual(j);
  }

  /* Then, check if any guys remain to initialize (outside the power-of-2-*/
  /* sized block), and if so, initialize them randomly. */
  if(stopping_guy_index != popsize) {
    starting_guy_index = stopping_guy_index ;
    init_random_pop(starting_guy_index);
  }
}


/* Initialize a population at random for a permutation-type problem */
void
initpermpop(starting_guy_index)
     int starting_guy_index;
{
  int     shuffled[MAXNUMFIELDS], notusedyet[MAXNUMFIELDS];
  int     i, j, k, start, stop, last;


  for (j = starting_guy_index; j < popsize; j++) {
    /* if all were read from file, loop not executed at all */
    for (k = 0; k < numfields; k++)
      notusedyet[k] = k;        /* one copy of each city, ready to pick from */
    last = numfields - 1;
    for (k = 0; k < numfields; k++) {
      i = rnd(0, last);
      shuffled[k] = notusedyet[i];
      notusedyet[i] = notusedyet[last--];
    }

    /* now write fields to their places in chromosome */
    start = 1;
    stop = fieldlength;
    for (i = 0; i < numfields; i++) {
      int2ithruj(start, stop, shuffled[i], oldpop[j].chrom);
      start += fieldlength;
      stop += fieldlength;
    }
    Initialize_For_New_Individual(j);
  }
}


/* When individuals are created randomly at initialization time, rather */
/* than being created as a result of genetic operators, values normally */
/* computed in generate() must be copied into the oldpop individuals here. */
void
Initialize_For_New_Individual(j)
     int j;
{
  /* For both initialization and for random reinitialization */

  oldpop[j].parent[0] = 0;      /* Initialize parent info. */
  oldpop[j].parent[1] = 0;
  oldpop[j].xsite[0] = 0;
  oldpop[j].xsite[1] = 0;
  remap_then_objfunc(&(oldpop[j])); /* Evaluate initial fitness */
}

void
initialize_inversion_map(map)
     /* initialize inversion map to identity map */
     int *map;
{
  int i;

  for(i=0;i<numfields;i++)
    map[i] = i;
  }


void
initreport_parallel()
     /* Initial report */
{
  int     indnum, i, j, numrows;
  int     printconvergence;     /* Boolean flag, really */
  struct individual *pind;

  skip(outfp, 1);
  fprintf(outfp, " GALOPPS Parameters for Subpopulation %d\n", popno);
  fprintf(outfp, " -------------------------------------------------\n");
  fprintf(outfp, " Using selection from %s\n", which_selection);
  fprintf(outfp, " Using crossover from %s\n", which_crossover);
  fprintf(outfp, " Using mutation from  %s\n", which_mutation);
  fprintf(outfp, " Using migration information from %s\n", masterfilename);
  fprintf(outfp, " Fitness function is %sstochastic\n", stochastic?"":"not ");
  if (permproblem) {
    fprintf(outfp, " Permutation-type problem, rearranging fields on chrom.\n");
    fprintf(outfp, "     (Therefore, inversion not permitted.)\n");
  }
  else {
    fprintf(outfp, " Not a permutation problem.\n"); 
    fprintf(outfp, " Using inversion from %s\n", which_inversion);
    fprintf(outfp, " Flag using_inversion is %s", using_inversion?"on":"off");
    if (using_inversion) {
      fprintf(outfp, ", so different inversion patterns\n   may be present,");
      fprintf(outfp, " even though inversion rate may sometimes be 0.\n");
    }
    else
      fprintf(outfp,", so inversion rate can NEVER\n   be > 0. in this run.\n");
  }
  fprintf(outfp, " User is %ssupplying own initialization\n"
		   , user_supplied_initialization?"":"not ");
  fprintf(outfp, " User is %ssupplying own migration code\n"
		  , call_app_user_does_migration?"":"not ");
  fprintf(outfp, " Different subpopulations using %s representation%s\n\n"
		   , different_reps? "different":"the same"
		   , different_reps? "s":" ");

  fprintf(outfp, " Number of cycles                   =   %d\n", ncycles);

  fprintf(outfp, " Total Population size              =   %d\n", popsize);
  fprintf(outfp," Generations per cycle              =   %d\n", genspercycle);
  fprintf(outfp, " Total number of fields             =   %d\n"
	  , numfields);
  if (alpha_size > 2) {
    fprintf(outfp, " Number of different codes/field    =   %d\n"
	    , alpha_size);
    fprintf(outfp, " Note:  crossover only between fields;\n");
    fprintf(outfp, " mutation only within alphabet.\n");
  }
  else if (alpha_size == 2)
    fprintf(outfp, " Binary chromosome, crossover anywhere\n");
  else {
    fprintf(outfp, " Using fields with different alphabet sizes, determined in\n");
    fprintf(outfp, " array field_sizes in function app_set_field_sizes in app file,\n");
    fprintf(outfp, " as follows:\n");
    numrows = numfields/10;
    for (j=0;j<numrows;j++)
      for (i=0;i<10 && j*10+i<numfields;i++)
	fprintf(outfp, "   %d", field_sizes[10*j+i]);
      fprintf(outfp,"\n");
  }
  fprintf(outfp, " Chromosome length (lchrom)         =   %d\n", lchrom);
  fprintf(outfp, " Maximum gen. number before ckpoint =   %d\n"
	  , gen + genspercycle);
  fprintf(outfp, " Crossover probability (pcross)     = %f\n", pcross);
  if (which_mutation[0] == 'm') {
    fprintf(outfp, " Mutation  probability (p_multimut)  = %f", pmutation);
    fprintf(outfp, " per individual\n");
    fprintf(outfp, " Average width (number of fields) of mutation = %f\n"
			, avg_mut_width);
  }                                                      
  else {
    fprintf(outfp, " Mutation  probability (pmutation)  = %f", pmutation);
    if (permproblem)
      fprintf(outfp, " per individual\n");
    else if (alpha_size == 2)
    fprintf(outfp, " per bit\n");
    else  
      fprintf(outfp, " per field\n");
  }
  if (!permproblem)
    fprintf(outfp, " Inversion probability (pinversion) = %f\n", pinversion);
  if (autocontrol_selection) {
    fprintf(outfp," Using automatically controlled\n   selection");
    fprintf(outfp," with Boltzmann scaling.\n");
    fprintf(outfp," Beta_s_tilde                       = %f\n", beta_s_tilde);
 
  }
  else if (beta >= 0.) {
      fprintf(outfp, " Using Boltzmann scaling, but\n");
      fprintf(outfp, " selection not automatically controlled.\n");
      fprintf(outfp, " Beta                               = %f\n", beta);
  }
  else {
    fprintf(outfp, " Not using autocontrolled selection or");
    fprintf(outfp, " Boltzmann scaling\n");
    fprintf(outfp, " Scaling Window (scaling_window)    =   %d\n"
	    , scaling_window);
    fprintf(outfp, " Coefficient for sigma truncation   = %f\n", sigma_trunc);
    fprintf(outfp, " Linear Scaling mult. (best/mean)   = %f\n", scalemult);
  }
  fprintf(outfp, " Crowding factor                    =   %d\n"
	  , crowding_factor);
  fprintf(outfp, " Incest Reduction                   =   %d\n"
	  , incest_reduction);
  fprintf(outfp, " S.D.'s above mean to be 'good'     = %f\n"
	  , conv_sigma_coeff);
  fprintf(outfp, " Interval betw. full converg. repts =   %d\n"
	  , convinterval);
  if (which_selection[0] == 't') 
    fprintf(outfp, " Tournament size                    = %d\n"
	    , tourneysize);
  fprintf(outfp, " Random seed                        = %f\n", randomseed);
  if (using_inversion) {
    fprintf(outfp, "\n* Current Inversion Map for Subpopulation %d:\n", popno);
    for (i=0;i<((numfields-1)/16) + 1;i++) {
      for (j=0;j < 16 && i * 16 + j < numfields;j++) {
	fprintf(outfp, "%d ", this_pop_map_to_std[16*i+j]);
      }
      fprintf(outfp, "\n");
    }
  }
  skip(outfp, 1);

  repchar(outfp, "-", LINELENGTH);
  skip(outfp, 1);
  if (printstrings == 1) {
    repchar(outfp, " ", ((LINELENGTH - 17) / 2));
    fprintf(outfp, "Population Report\n");
    fprintf(outfp, "Generation %3d:\n", gen);
    fprintf(outfp, "Individual number\n");
    fprintf(outfp, "binary representation of chromosome\n");
    if (alpha_size != 2)
	fprintf(outfp, "chromosome as int fields\n");
    fprintf(outfp, "raw fitness | scaled fitness\n");
    repchar(outfp, "-", LINELENGTH);
    skip(outfp, 1);

    writeinitpop();

    skip(outfp, 1);

  }

  fprintf(outfp, "              Initial Population Summary Report\n");
  fprintf(outfp,"Raw fitnesses:  min = %f, max = %f,\n                avg = %f",
	  minraw, oldpop[bestnow].init_fitness, avgraw);
  printconvergence = conv_sigma_coeff != 7 && convinterval != 0
    && gen % convinterval == 0 && quiet == 0;
  if (printconvergence || sigma_trunc != 0.)
    fprintf(outfp, ", sigma= %f", sigma);
  if (beta >= 0.)
    fprintf(outfp, "\nBoltzmann's beta = %f\n", beta);
  fprintf(outfp, "\n");
  fprintf(outfp, "Scaled fitnesses:  min = %g, max = %g,", fit_min, fit_max);
  fprintf(outfp, "\n                   avg = %g, sum = %f\n"
	  , fit_avg, sumfitness);
  if (convinterval != 0) {
    fprintf(outfp, "Convergence:  %6.2f%% alleles lost, ", percent_lost);
    fprintf(outfp, "%6.2f%% converged, ", percent_converged);
    fprintf(outfp, "avg. bias %6.2f%%\n", avg_bias);
  }
  indnum = bestnow + 1;
  fprintf(outfp, "Best Individual of Initial Population: Indiv. #: %d\n"
	  ,indnum);             /* add one for external numbering */
  pind = &(oldpop[bestnow]);
  writechrom(pind->chrom);
  if (alpha_size != 2)  
    writefields(pind->chrom);
  skip(outfp, 1);

  if (printconvergence)
    reptconv();

  /* application dependent report */
  app_initreport();

  fflush(outfp);
}

void
writeinitpop()
{
  struct individual *pind;
  int     j;

  for (j = 0; j < popsize; j++) {
    fprintf(outfp, "%3d) \n", j + 1);

    /* Old string */
    pind = &(oldpop[j]);
    writechrom(pind->chrom);
    if (alpha_size != 2)
	writefields(pind->chrom);
    fprintf(outfp, "\n  %8f | %8f\n", pind->init_fitness, pind->fitness);
  }
}
