/*----------------------------------------------------------------------------*/
/* mainone.c -- main program for the single-population version of the         */
/*              GALOPP System.                                                */
/*----------------------------------------------------------------------------*/

/************************************************/
/* GALOPPS:  The "Genetic ALgorithm Optimized   */
/*   for Portability and Parallelism" System    */
/*                                              */
/* (c) 1993,1994,1995, 1996 Board of Trustees         */ 
/*          Michigan State University           */
/*       Written by Erik David Goodman          */
/*        Genetic Algorithms Research           */
/*          and Applications Group              */
/* Michigan State University, East Lansing, MI  */
/*                                              */
/* THIS IS FREE SOFTWARE, AND IS DISTRIBUTED    */
/* ACCORDING TO THE TERMS OF THE GNU PUBLIC     */
/* LICENSE, LATEST VERSION.  INFORMATION ABOUT  */
/* ITS COPYING AND USE IS CONTAINED IN THE FILE */
/* 'license' IN THE 'docs' SUBDIRECTORY WHICH   */
/* ACCOMPANIES THIS DISTRIBUTION.  THERE IS     */
/* ABSOLUTELY NO WARRANTY, EXPRESSED OR IMPLIED,*/
/* ASSOCIATED WITH THIS SOFTWARE, AND ALL RISK  */
/* FOR ITS USE OR DISTRIBUTION RESTS WITH THE   */
/* USER.                                        */
/*                Release 3.2                   */
/*                                              */
/*          Single Population Module            */
/*                                              */
/*          A Distant Derivative of             */
/*       Simple Genetic Algorithm - SGA         */
/*              Haploid Version                 */
/* (c)    David Edward Goldberg  1986           */
/*            All Rights Reserved               */
/*        C translation by R.E. Smith           */
/*    v1.1 modifications by Jeff Earickson,     */
/*    From which the ESGA/IPGA System was       */
/*               developed by                   */
/*  Erik Goodman, Michigan State University     */
/* & Wang Gang, Beijing U. of Aero. & Astro.,   */
/*  which, after 3 releases, led to GALOPPS.    */
/************************************************/
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "sga.h"

#include <time.h>
#ifdef __MWERKS__
  #include <console.h>
#endif

#ifdef PROTOTYPES_ACCEPTED
int commandline(int, char**);
#else
int commandline();
#endif
void input_overrides();
void global_overrides();
void plot();

main(argc, argv)
int     argc;
char   *argv[];
{
  struct individual *temp;
  char   c;
  FILE   *fopen();
  void    copyright();
  
  /* one process. */
  many = 0;
  maxruns = 1;
  single_process = TRUE;
  startgen = 0;       /* initial generation number for non-restart runs */
  gen = 0;
  quiet = 0;

  /* Process the command line... */
  #ifdef __MWERKS__
    argc = ccommand( &argv );
  #endif
  /* Determine input and output from program args */
  parameter_changes = commandline(argc, argv);
  
  /* print the author/copyright notice */
  copyright(0);
  
  if (interactive) 
    printf("Number of GA runs to be performed (%d) -> ", maxruns);
  ffscanf("numberofruns", infp, "%d", &maxruns);
  
  for (run = 1; run <= maxruns; run++) {
    if (quiet < 3)
      fprintf(outfp, "\n*** Beginning of run %d ***\n", run);

    /* Get configuration inputs for this run */
    GetThisRunInputs();

    for (gen = startgen; gen < maxgen; gen++) {
      if (quiet == 0) 
	fprintf(outfp, "\nRUN %d of %d: GENERATION %d->%d\n",
		run, maxruns, gen, maxgen);

      /* application dependent routines */
      application();

      /* create a new generation */
      generation();

      /* compute fitness statistics on new populations */
      statistics(newpop);
      windowend++;              /* advance window for window scaling */

      /* report results for new generation */
      plot();
      if (quiet < 3)
	report();
      if (quiet > 0)
	/* zeroes for parameters below used in GALOPPS/Manypops, */
	/* not Onepop version. */
	app_quiet_report();

      /* advance the generation */
      temp = oldpop;
      oldpop = newpop;
      newpop = temp;

      /* Allow user to make changes if population performance or */
      /* convergence measures meet certain user-specified criteria.  */
      ModifyRunBasedOnConvergence();

      if (ckptfreq > 0 && gen % ckptfreq == 0)
	write_checkpoint();
	update_rstate(checkpthdrfilename,1);
	update_rstate(checkptindfilename,1);
    }

    /* in case didn't run any generations, fix up statistics before */
    /* write out the checkpoint file.                               */
    if (startgen >= maxgen)
      statistics(oldpop);

    write_checkpoint();
    update_rstate(checkpthdrfilename,1);
    update_rstate(checkptindfilename,1);
    fflush(outfp);
    freeall();
  }

  return;
}

/* Query user (or input file) for the parameters which will stay the */
/* same over all of the runs requested (one or more). */
void
GetThisRunInputs()
{
  if (interactive && run > 1)
    printf("\n\nData entry for run %d:\n\n",run);
  
  GetUniversalInputs();
  
  if (!extending) {
    problem_inputs();

    if (interactive)
      printf("Maximum number of generations -> ");
    ffscanf("maxgen", infp, "%d", &maxgen);
  }

  /* Call app_set_options() to let user set conditions for run.  */
  app_set_options();

  /* Get global command-line overrides, if any (unless interactive) */
  if (!interactive)
    global_overrides();

  /* Set things up */
  initialize();

  if (permproblem)
    /* Malloc temporary space needed for permutation-type operators */
    MallocPermOperSpace();
}


void
ModifyRunBasedOnConvergence()
/* Allow user chance to decide converged, replace some or all */
/* before advancing generation, or quit, or whatever wants. */
{
    static int converg_cond_flag; /* flag used in app routine for reporting */
				  /* results of test in                     */
				  /* app_decide_if_converged().             */
				  /* if 0, continue run, no change */
				  /* if 1, reinitialize population, starting */
				  /*     from guy "starting_guy_index        */
				  /* if 2, call app_when_converged() to allow */
				  /* user to take whatever action desired.    */
    static int starting_guy_index; /* index of first indiv to initialize at */
				   /* random. */
    static int number_to_replace; /* On reinitialize, tells how many to do. */
    int i, newbest;

    converg_cond_flag = 0;
    app_decide_if_converged(&converg_cond_flag, &number_to_replace);
    switch (converg_cond_flag) {
    case 0:    /* Just keep going */
		break;
    case 1:    /* keep "starting_guy_index" guys, randomly reinitialize */
		       /* all the rest. */
	starting_guy_index = popsize - number_to_replace;

	if (quiet < 3) {
	  fprintf(outfp, "\n *** Population converged ***");
	  fprintf(outfp, "\n Reinitializing %d individuals.\n"
		  , number_to_replace);
	}
	if (elitism && starting_guy_index > 0) {
	  /* Be sure last copy of best isn't overwritten -- store him in */
	  /* slot 0, reinitialize later guys */
	  oldpop[0] = oldpop[bestnow];
	  bestnow = 0;
	}

	if (user_supplied_initialization)
	    app_user_init_pop(starting_guy_index);
	else if (permproblem)
	    initpermpop(starting_guy_index);
	else if(alpha_size != 2)
	    init_non_binary(starting_guy_index);
	else if(superuniform_initialization)
	    init_superuniform_pop(starting_guy_index);
	else
	    init_random_pop(starting_guy_index);

	/* Record NEW best guy, whether elitism on or not. */
	newbest = bestnow;
	for (i=0;i<popsize;i++) {
	  if (oldpop[i].init_fitness > oldpop[bestnow].init_fitness)
	    newbest = i;
	}
	if (newbest != bestnow) {
	  bestnow = newbest;
	  bestfit.init_fitness = oldpop[bestnow].init_fitness;
	  bestfit.neval = oldpop[bestnow].neval;
	  bestfit.generation = gen;
	  for (i=0;i<chromsize;i++)
	    bestfit.chrom[i] = oldpop[bestnow].chrom[i];
	  sys_util_copy_utility(bestfit.utility, oldpop[bestnow].utility);
	}
	break;

    case 2:
	app_when_converged();
	break;
    default:
	fprintf(stderr, "\n *** Logic error in mainone.c or app code ***.\n");
	exit(-111);
    }
    return;
}
