/*----------------------------------------------------------------------------*/
/* startup.c - functions to get things set up and initialized for a new run or*/
/*             for a restart from a checkpoint file.  Allows resetting        */
/*             of parameters by user to values different from what were in    */
/*             restart file, or defaulting of all but maxgen to values used   */
/*             when file was run.  In all cases, at restart time, the default */
/*             values presented to the user are the values from the saved     */
/*             checkpoint file.  Checkpt file ALWAYS written.                 */
/*----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include "external.h"

static int numberwantfromfile;  /* no. of indiv's to be read from .ind file */


/* Initialization Coordinator */
void
initialize()
{
  unsigned  nbytes;
  char   *v;
  int     i;              /* index */
  int     indcnt;         /* number of indiv's available in .ind file  */
  char  answer = 'y';
  BOOL    Must_Recalculate_Bestnow = FALSE;
  static int starting_guy_index; /* index of first guy to initialize or */
				    /* reinitialize to random value */

  strcpy(checkpthdrfilename, checkptfileprefix);
  strcat(checkpthdrfilename, ".ckp");
  strcpy(checkptindfilename, checkptfileprefix);
  strcat(checkptindfilename, ".ind");

  /* Here is logic for restarting from saved checkpoint. */
  if (strlen(restartfileprefix)) {
    strcpy(restarthdrfilename, restartfileprefix);
    strcat(restarthdrfilename, ".ckp");

    /* get default parameter values from checkpoint file */
    if ((indcnt = ReadCheckPointHeader(restarthdrfilename, app_read_ckp_hdr))
	 == -1) {
      /* Read failed. */
      fprintf(stderr, "*** Exiting because can't read restart");
      fprintf(stderr, " header file %s.\n", restarthdrfilename);
      exit(-1);
    }

    startgen = gen;

    if (interactive)
      fprintf(outfp, "Enter new maxgen ----------------------->");
    ffscanf("maxgen", infp, "%d", &maxgen);
    if (maxgen < startgen)
      maxgen = startgen;

    /* Now, extending, if NOT interactive, must re-install override on */
    /* maxgen since was just read from input file, and may be overridden */
    if (parameter_changes && !interactive)
      if ((v = get_parameter("maxgen", 0))) {
	maxgen = atoi(v);
      }

    if (interactive)
      fprintf(outfp, "Change parameters besides maxgen?([y]/n)-->");

    ffscanf("other_changes", infp, "%c", &answer);
    if (answer != 'n' && answer != 'N')
      /* allow revision of parameters with defaults from file */
      pop_inputs(&Must_Recalculate_Bestnow);

    /* Even if didn't make changes in input file, may be command-line */
    /* overrides to install here, unless interactive: */
    if (!interactive)
      get_pop_overrides();

    /* Continue with initialization */

    /* malloc population and some other variables, */
    initmalloc();

    /* malloc utility fields for newpop and oldpop populations */
    sys_util_malloc();

    numberwantfromfile = popsize > indcnt ? indcnt : popsize; /* min */

    strcpy(restartindfilename, restartfileprefix);
    strcat(restartindfilename, ".ind");
    read_checkpointfile(restartindfilename, numberwantfromfile
			, Must_Recalculate_Bestnow);

    /* give application a chance to do its thing, must figure out */
    /* that it is a restart by itself  */
    app_init();

    /* call a revised initpop here; it will create popsize -    */
    /* numberwantfromfile individuals, and stick them on end,   */
    /* and NOT set neval to popsize, but add new eval # to it,  */
    starting_guy_index = numberwantfromfile;
    if (user_supplied_initialization)
      app_user_init_pop(starting_guy_index);
    else if (permproblem)
      initpermpop(starting_guy_index);
    else if(alpha_size != 2)
      init_non_binary(starting_guy_index);
    else if(superuniform_initialization)
      init_superuniform_pop(starting_guy_index);
    else
      init_random_pop(starting_guy_index);

    /* Let user clean up the random initialization here. */
    app_after_random_init();

    statistics(oldpop);

    if (quiet <3)
      initreport();
  }

  else {
    /* Now handling the case NOT extending, start from beginning. */
    numberwantfromfile = 0;

    /* get basic problem values from input file */
    initdata();

    /* Now know numfields, chromsize, so can set max variables for */
    /* compatibility with Manypops. */
    max_chromsize_among_all_subpops = chromsize;
    max_numfields_among_all_subpops = numfields;

    /* Malloc the populations and various fixed-length structures */
    initmalloc();

    /* Malloc below is handled as a special case, because it must NOT */
    /* be done in initmalloc() if restarting from checkptfile (then   */
    /* it MUST be done inside ReadCheckPointHeader).  Therefore, it is  */
    /* done separately here, for the case NOT restarting from a saved */
    /* checkpoint file */
    nbytes = numfields * sizeof(int);
    if ((this_pop_map_to_std = (int *) malloc(nbytes)) == NULL)
      nomemory("this_pop_map_to_std");

    /* Malloc the utility fields for the newpop and oldpop populations */
    sys_util_malloc();

    /* Malloc the variable-length fields for the bestever individual */
    /* variables (for checkpoint file compatibility with Manypops) */
    bestever_variable_malloc();

    /* initialize application dependent variables */
    app_init();

    /* initialize random number generator */
    Randomize(&randomseed);

    /* initialize global counters/values */
    nmutation = 0;
    ncross = 0;
    neval = 0;
    bestfit.fitness = -1.e37;
    bestfit.init_fitness = -1.e37;
    bestfit.neval = 0;
    bestfit.generation = 0;
    windowstart = 0;
    windowend = 0;

    /* Even if not using inversion, initialize map to identity.     */
    /* This allows shared file format for checkpoint header files   */
    /* between Onepop and Manypops, with & without inversion.       */
    initialize_inversion_map(this_pop_map_to_std);

    /* initialize the populations and report statistics */
    starting_guy_index = 0;
    if (user_supplied_initialization)
      app_user_init_pop(starting_guy_index);
    else if (permproblem)
      initpermpop(starting_guy_index);
    else if (alpha_size != 2)
      init_non_binary(starting_guy_index);
    else if (superuniform_initialization)
      init_superuniform_pop(starting_guy_index);
    else
      init_random_pop(starting_guy_index);

    /* Let user clean up or redo or the random initialization here. */
    app_after_random_init();

    statistics(oldpop);
    windowend++;
    if (quiet < 3)
      initreport();
  }
}


/* data inquiry and setup at start of a run. */
void
initdata()
{
  extern int parameter_changes;
  char    answer;
  BOOL f = FALSE;

  static  BOOL firstcall = TRUE;
  int   nbytes, i, k, fldsiz, fldlen, prev_end;
  int   *inverse_map_to_std;

  if (firstcall) {
    firstcall = FALSE;
    pcross = -1.0;
    pmutation = -1.0;
    pinversion = 0.0;
    scaling_window = -1;
    sigma_trunc = 0.0;
    scalemult = 0;
    crowding_factor = 0;
    incest_reduction = 0;
    conv_sigma_coeff = 7.0;
    convinterval = 10;
    popno = 0;
  }

  if (interactive )
    fprintf(outfp, "\n ------- Onepop Data Entry and Initialization -------\n");

  pop_inputs(&f);
  if (parameter_changes && !interactive)
    get_pop_overrides();

    
  /* Now, regardless of alpha_size, get needed arrays malloc'd  (some */  
  /* things are done here only for compatibility with Manypops, since */
  /* a Onepop checkpoint file might be used as part of a Manypops */
  /* restart (extension) by some users, and since they use the same */
  /* checkpoint read/write routines. */

  /* If max sizes for bestever variables was not set by user in function */
  /* app_set_options, do something here which should be correct. */
  if (max_numfields_among_all_subpops == -1)
    max_numfields_among_all_subpops = numfields;
  
  malloc_field_sizes_arrays();          

  if (alpha_size >= 2) {
    /* Set up array field_sizes */
    for (i=0;i<numfields;i++) {
      field_sizes[i] = alpha_size;
    }
    /* Since field_sizes all equal, get ceiling log base 2 of alpha_size for */
    /* fieldlength, use to set field_ends[].  */
    k = 1;
    for (fieldlength = 1; fieldlength < 17; fieldlength++) {
      k *= 2;
      if (k >= alpha_size)
	break;
    }
    field_ends[0] = fieldlength;
    for (i=1;i<numfields;i++) {
      field_ends[i] = field_ends[i-1] + fieldlength;
    }
  }
    
  /* Else, must get length (bits) of each field specified in field_sizes[] */
  else {
    /* Set to illegal values, so can be sure user set them. */
    for (i=0;i<numfields;i++)
      field_sizes[i] = -1;
    app_set_field_sizes();
    /* Now be sure user set all of them to legal values */
    for (i=0;i<numfields;i++)
      if (field_sizes[i] < 2) {
	fprintf(stderr, "*** ERROR.  User set alpha_size < 2, indicating");
	fprintf(stderr, " that field_sizes are\ndifferent from field to ");
	fprintf(stderr, "field.  However, user did NOT specify legal ");
	fprintf(stderr, "field sizes\nfor all %d fields in ", numfields);
	fprintf(stderr, "function app_set_field_sizes() in the ");
	fprintf(stderr, "app-specific\nfile.  Run cannot continue.  ***\n");
	fprintf(outfp, "*** ERROR.  User set alpha_size < 2, indicating");
	fprintf(outfp, " that field_sizes are\ndifferent from field to ");
	fprintf(outfp, "field.  However, user did NOT specify legal ");
	fprintf(outfp, "field sizes\nfor all %d fields in ", numfields);
	fprintf(outfp, "function app_set_field_sizes() in the ");
	fprintf(outfp, "app-specific\nfile.  Run cannot continue.  ***\n");
	exit (-1);
      }
    prev_end = 0;
    for (i=0;i<numfields;i++) {
      fldsiz = field_sizes[i];
      k = 1;
      for (fldlen = 1; fldlen < 17; fldlen++) {
	k *= 2;
	if (k >= fldsiz)
	  break;
      }
      field_ends[i] = prev_end + fldlen;
      prev_end += fldlen;
    }
  }

  /* Malloc below is handled as a special case, because it must NOT */
  /* be done in initmalloc() if restarting from checkptfile (then   */
  /* it MUST be done inside ReadCheckPointHeader).  Therefore, it is  */
  /* done separately here, for the case NOT restarting from a saved */
  /* checkpoint file */
  nbytes = numfields * sizeof(int);
  if ((this_pop_map_to_std = (int *) malloc(nbytes)) == NULL)
    nomemory("this_pop_map_to_std");
    
  /* Even though not using inversion, initialize map to identity. */

  initialize_inversion_map(this_pop_map_to_std);
    
  firstcall = FALSE;
  /* Create this info for Standard population, in case later */
  /* is changed by inversion, etc.  Might not be the rep of any subpop */
	
  /* Allocate temporary memory for inverse of map_to_std array */
  nbytes = numfields * sizeof(int);
  if ((inverse_map_to_std = (int *) malloc(nbytes)) == NULL)
     nomemory("inverse_map_to_std");
  /* Create inverse map to std using temporary array just malloc'd */
  for (i=0;i<numfields;i++) 
    inverse_map_to_std[this_pop_map_to_std[i]] = i; 

  /* Now, use that to calculate std_field_ends, the field_ends array */
  /* for use with the STANDARD representation */
  for (i=0;i<numfields;i++) {
    std_field_ends[i] = (i==0 ? 0 : std_field_ends[i-1]) 
		      + (field_ends[inverse_map_to_std[i]] 
		      - (i==0 ? 0 : field_ends[inverse_map_to_std[i]-1]));
  }
  free(inverse_map_to_std);
  inverse_map_to_std = NULL;

  lchrom = field_ends[numfields - 1];
  /* define chromosome size in terms of machine bytes, ie  */
  /* length of chromosome in bits (lchrom)/(bits-per-byte) */
  /* chromsize must be known for malloc() of chrom pointer */
  chromsize = (lchrom / UINTSIZE);
  if (lchrom % UINTSIZE)
    chromsize++;
  if (interactive)
    printf("   * Each chromosome will be %d bit(s), %d unsigned(s) long.\n", 
	   lchrom, chromsize);
  if (max_chromsize_among_all_subpops == -1)
    max_chromsize_among_all_subpops = chromsize;

}


/* Initialize a population at random */
void
init_random_pop(starting_guy_index)
     int starting_guy_index;
{
  int     j, j1, k, stop;
  unsigned mask = 1;

  for (j = starting_guy_index; j < popsize; j++) {
    for (k = 0; k < chromsize; k++) {
      oldpop[j].chrom[k] = 0;
      if (k == (chromsize - 1))
	stop = lchrom - (k * UINTSIZE);
      else
	stop = UINTSIZE;

      /* A fair coin toss */
      for (j1 = 1; j1 < stop; j1++) { /* EDG fix makes <= a < */
	if (flip(0.5))
	  oldpop[j].chrom[k] = oldpop[j].chrom[k] | mask;
	oldpop[j].chrom[k] = oldpop[j].chrom[k] << 1;
      }
      /* handle least significant bit in each word (leftmost in    */
      /* printouts), without shifting. (Bug fixed by EDG Oct 02,93. */
      if (flip(0.5))
	oldpop[j].chrom[k] = oldpop[j].chrom[k] | mask;

    }
    Initialize_For_New_Individual(j);
  }
}


/* Initialize a population at random for a permutation-type problem */
void
initpermpop(starting_guy_index)
     int starting_guy_index;
{
  int     shuffled[MAXNUMFIELDS], notusedyet[MAXNUMFIELDS];
  int     i, j, k, start, stop, last;


  for (j = starting_guy_index; j < popsize; j++) {
    for (k = 0; k < numfields; k++)
      notusedyet[k] = k;        /* one copy of each poss. value, ready to
				 * pick from */
    last = numfields - 1;
    for (k = 0; k < numfields; k++) {
      i = rnd(0, last);
      shuffled[k] = notusedyet[i];
      notusedyet[i] = notusedyet[last--];
    }

    /* now write fields to their places in chromosome */
    start = 1;
    stop = fieldlength;
    for (i = 0; i < numfields; i++) {
      int2ithruj(start, stop, shuffled[i], oldpop[j].chrom);
      start += fieldlength;
      stop += fieldlength;
    }
    Initialize_For_New_Individual(j);
  }
}


/* Initialize a population at random for non-binary, */
/* non-permutation-type problem (alpha_size != 2, not nec. a pwr of 2 */
void
init_non_binary(starting_guy_index)
     int starting_guy_index;
{
  int     i, j, start, randvalue;


  for (j = starting_guy_index; j < popsize; j++) {

    /* write random fields (in range) to their places in chromosome */
    start = 1;
    for (i = 0; i < numfields; i++) {
      randvalue = rnd(0, field_sizes[i] - 1);
      int2ithruj(start, field_ends[i], randvalue, oldpop[j].chrom);
      start = field_ends[i] + 1;
    }
    Initialize_For_New_Individual(j);
  }
}


/* Perform superuniform initialization of a population of BINARY chromosomes, */
/* starting at population index starting_guy_index (lower ones were read from */
/* file) and continuing for the largest power of 2 remaining below popsize.   */
/* For any remaining guys, initialize them at random.  Note:  OPTIONAL        */
void
init_superuniform_pop(starting_guy_index)
     int starting_guy_index;
{
  int     j, k, n_superuniform_initialize, stopping_guy_index;
  static    int     num_superuniform_cycles, bits_left_over, startvalue
    , cycle_index;
  int     num_patterns_with_bits_left_over, startbit, stopbit;

  superuniform_cyclebitwidth = floorlogbase2(popsize - starting_guy_index) ;
  n_superuniform_initialize = (int)powerof2(superuniform_cyclebitwidth) ;
  stopping_guy_index = starting_guy_index + n_superuniform_initialize;
  num_superuniform_cycles = lchrom/superuniform_cyclebitwidth; /* floor */
  bits_left_over = lchrom % superuniform_cyclebitwidth; /* special case  */
  num_patterns_with_bits_left_over = (int)powerof2(bits_left_over);

  startbit = 1;
  stopbit = superuniform_cyclebitwidth;
  for (cycle_index=0; cycle_index< num_superuniform_cycles; cycle_index++) {

    /* Pick a random value in the range to start the sequence at.  */
    startvalue = rnd(0, (int)(powerof2(superuniform_cyclebitwidth) - 1));

    /* Write values for this field to their places on the chromosomes */
    for (j = starting_guy_index; j < stopping_guy_index; j++) {
      k = (j + startvalue) % n_superuniform_initialize;
      int2ithruj(startbit, stopbit, k, oldpop[j].chrom);
    }

    startbit += superuniform_cyclebitwidth;
    stopbit += superuniform_cyclebitwidth;
  }

  /* Now handle the final partial-width field (if any) */
  stopbit = startbit + bits_left_over - 1;
  startvalue = rnd(0, num_patterns_with_bits_left_over - 1);
  for (j = starting_guy_index; j < stopping_guy_index; j++) {
    if(stopbit >= startbit) {
      k = (j + startvalue) % num_patterns_with_bits_left_over;
      int2ithruj(startbit, stopbit, k, oldpop[j].chrom);
    }
    /* and fill in the rest of the fields, while we're at it. */
    Initialize_For_New_Individual(j);
  }
  /* Then, check if any guys remain to initialize (outside the power-of-2-*/
  /* sized block), and if so, initialize them randomly. */
  if(stopping_guy_index != popsize) {
    starting_guy_index = stopping_guy_index ;
    init_random_pop(starting_guy_index);
  }
}


/* When individuals are created randomly at initialization time, rather */
/* than being created as a result of genetic operators, values normally */
/* computed in generate() must be copied into the oldpop individuals here. */
void
Initialize_For_New_Individual(j)
     int j;
{
  /* For both initialization and for random reinitialization */

  oldpop[j].parent[0] = 0;      /* Initialize parent info. */
  oldpop[j].parent[1] = 0;
  oldpop[j].xsite[0] = 0;
  oldpop[j].xsite[1] = 0;
  remap_then_objfunc(&(oldpop[j])); /* Evaluate initial fitness */
}


void
initialize_inversion_map(map)
     /* initialize inversion map to identity map */
     int *map;
{
  int i;

  for(i=0;i<numfields;i++)
    map[i] = i;
}


/* Output report on initialized population */
void
initreport()
{
  int     indnum, i, j, numrows;

  skip(outfp, 1);
  fprintf(outfp, " GALOPPS Parameters\n");
  fprintf(outfp, " -------------------------------------------------\n");
  fprintf(outfp, " Using selection from %s\n", which_selection);
  fprintf(outfp, " Using crossover from %s\n", which_crossover);
  fprintf(outfp, " Using mutation  from %s\n", which_mutation);
  fprintf(outfp, " Fitness function is %sstochastic\n", stochastic?" ":"not ");
  if (permproblem) {
    fprintf(outfp, " Permutation-type problem, rearranging fields on chrom.\n");
    fprintf(outfp, "     (Therefore, inversion not permitted.)\n");
  }
  else {
    fprintf(outfp, " Not a permutation problem.\n");
    fprintf(outfp, " Using inversion from %s\n", which_inversion);
    fprintf(outfp, " Flag using_inversion is %s", using_inversion?"on":"off");
    if (using_inversion) {
      fprintf(outfp, ", so different inversion patterns\n   may be present,");
      fprintf(outfp, " even though inversion rate may sometimes be 0.\n");
    }
    else
      fprintf(outfp,", so inversion rate can NEVER\n   be > 0. in this run.\n");
  }
  fprintf(outfp, " User is %ssupplying own initialization\n"
  	     , user_supplied_initialization?"":"not ");
  fprintf(outfp, " User is %ssupplying own migration code\n\n"
		  , call_app_user_does_migration?"":"not ");
  fprintf(outfp, " Total Population size              =   %d\n", popsize);

  fprintf(outfp, " Total number of fields             =   %d\n", numfields);
  if (alpha_size > 2) {
    fprintf(outfp, " Number of different codes/field    =   %d\n", alpha_size);
    fprintf(outfp, " Note:  crossover only between fields;\n");
    fprintf(outfp, " mutation only within alphabet.\n");
  }
  else if (alpha_size == 2)
    fprintf(outfp, " Binary chromosome, crossover anywhere\n");
  else {
    fprintf(outfp, " Using fields with different alphabet sizes, specified in\n"); 
    fprintf(outfp, " array field_sizes in function app_set_field_sizes in app file,\n");
    fprintf(outfp, " as follows:\n");
    numrows = numfields/10;
    for (j=0;j<numrows;j++)
      for (i=0;i<10 && j*10+i<numfields;i++)
	fprintf(outfp, "   %d", field_sizes[10*j+i]);  
      fprintf(outfp,"\n"); 
  }
  fprintf(outfp, " Chromosome length (lchrom)         =   %d\n", lchrom);
  fprintf(outfp, " Maximum # of generations (maxgen)  =   %d\n", maxgen);
  fprintf(outfp, " Crossover probability (pcross)     = %f\n", pcross);
  if (which_mutation[0] == 'm') {
    fprintf(outfp, " Mutation  probability (p_multimut)  = %f", pmutation);
    fprintf(outfp, " per individual\n");
    fprintf(outfp, " Average width (number of fields) of mutation = %f\n"
	    , avg_mut_width);
  }
  else {
    fprintf(outfp, " Mutation  probability (pmutation)  = %f", pmutation);
    if (permproblem)
      fprintf(outfp, " per individual\n");
    else if (alpha_size == 2)
      fprintf(outfp, " per bit\n");
    else  
      fprintf(outfp, " per field\n");
  }
  if (!permproblem)
    fprintf(outfp, " Inversion probability (pinversion) = %f\n", pinversion);
  if (autocontrol_selection) {
    fprintf(outfp," Using automatically controlled\n   selection");
    fprintf(outfp, "with Boltzmann scaling.\n");
    fprintf(outfp," Beta_s_tilde                       = %f\n", beta_s_tilde);
  }
  else if (beta >= 0.) {
    fprintf(outfp, " Using Boltzmann scaling, but\n");
    fprintf(outfp, " selection is not automatically controlled.\n");
    fprintf(outfp, " Beta                               = %f\n", beta);
  }
  else {
    fprintf(outfp, " Not using autocontrolled selection or"); 
    fprintf(outfp, " Boltzmann scaling.\n");
    fprintf(outfp, " Scaling Window (scaling_window)    =   %d\n"
	    , scaling_window);
    fprintf(outfp, " Coefficient for sigma truncation   = %f\n", sigma_trunc);
    fprintf(outfp, " Linear Scaling mult. (best/mean)   = %f\n", scalemult);
  }
  fprintf(outfp, " Crowding factor                    =   %d\n"
	  , crowding_factor);
  fprintf(outfp, " Incest Reduction                   =   %d\n"
	  , incest_reduction);
  fprintf(outfp, " S.D.'s above mean to be 'good'     = %f\n"
	  , conv_sigma_coeff);
  fprintf(outfp, " Interval betw. full converg. repts =   %d\n"
	  , convinterval);
  if (which_selection[0] == 't')
    fprintf(outfp, " Tournament size                    = %d\n"
	    , tourneysize);
  fprintf(outfp, " Random Seed                        = %f\n", randomseed);
  if (using_inversion) {
    fprintf(outfp,"\n* Current Inversion Map for Subpopulation %d:\n",popno);
    for (i=0;i<((numfields-1)/16) + 1;i++) {
      for (j=0;j < 16 && i * 16 + j < numfields;j++) {
	fprintf(outfp, "%d ", this_pop_map_to_std[16*i+j]);
      }
      fprintf(outfp, "\n");
    }
  }
  skip(outfp, 1);

  if (printstrings == 1) {
    repchar(outfp, " ", ((LINELENGTH - 17) / 2));
    fprintf(outfp, "Population Report\n");
    fprintf(outfp, "Generation %3d:\n", gen);
    fprintf(outfp, "Individual number\n");
    fprintf(outfp, "binary representation of chromosome\n");
    if (alpha_size != 2)
	fprintf(outfp, "chromosome as int fields\n");
    fprintf(outfp, "raw fitness | scaled fitness\n");
    repchar(outfp, "-", LINELENGTH);
    skip(outfp, 1);

    writeinitpop();
  }
  fprintf(outfp, "              Initial Population Summary Report\n");
  fprintf(outfp, "Raw fitnesses:  min = %f, max = %f,\n", 
	  minraw, oldpop[bestnow].init_fitness);
  fprintf(outfp, "                avg = %f", avgraw);
  if ((conv_sigma_coeff != 7. && convinterval != 0 &&
       gen % convinterval == 0) || sigma_trunc != 0.)
    fprintf(outfp, ", sigma= %f", sigma);
  fprintf(outfp, "\n");
  if (beta >= 0.)
    fprintf(outfp, "Boltzmann's beta = %f\n", beta);
  fprintf(outfp, "Scaled fitnesses:  min = %g, max = %g,\n ", fit_min, fit_max);
  fprintf(outfp, "                  avg = %g, sum = %f\n", fit_avg, sumfitness);
  if (convinterval != 0) {
     fprintf(outfp, "Convergence:  %6.2f%% alleles lost, ", percent_lost);
     fprintf(outfp, "%6.2f%% converged, ", percent_converged);
     fprintf(outfp, "avg. bias %6.2f%%\n", avg_bias);
  }
  indnum = bestnow + 1;
  fprintf(outfp, "Best Individual of Initial Population: Indiv. #: %d\n"
	  ,indnum);             /* add one for external numbering */
  writechrom(oldpop[indnum].chrom);
  if (alpha_size != 2)
    writefields(oldpop[indnum].chrom);
  if (convinterval != 0 && quiet == 0)
    reptconv();

  /* application dependent report */
  app_initreport();

  fflush(outfp);
}

void
writeinitpop()
{
  struct individual *pind;
  int     j;

  for (j = 0; j < popsize; j++) {
    fprintf(outfp, "%3d) \n", j + 1);

    /* Old string */
    pind = &(oldpop[j]);
    writechrom(pind->chrom);
    if (alpha_size != 2) 
	writefields(pind->chrom);
    fprintf(outfp, "\n  %8f | %8f\n", pind->init_fitness, pind->fitness);
  }
}


/* Below is a dummy, never called in Onepop, to avoid a linker error.  */
BOOL
ReadNeighborTable(filename)
char    *filename;
{
  return TRUE;
}


